from ml.ml40.features.properties.values.value import Value


class Assortment(Value):
    def __init__(self, namespace="fml40", name="", identifier="", parent=None):
        super().__init__(
            namespace=namespace, name=name, identifier=identifier, parent=parent
        )
        self.__grade = None
        self.__json_out = dict()

    @property
    def grade(self):
        return self.__grade

    @grade.setter
    def grade(self, value):
        self.__grade = value

    def to_json(self):
        self.__json_out = super().to_json()
        if self.grade is not None:
            self.__json_out["grade"] = self.grade

        return self.__json_out



from ml.ml40.features.properties.values.value import Value


class WhiteLevel(Value):
    def __init__(self, namespace="ml40", name="", identifier="", parent=None):
        super().__init__(
            namespace=namespace,
            name=name,
            identifier=identifier,
            parent=parent)



from ml.ml40.features.properties.values.value import Value


class Displacement(Value):
    def __init__(self, namespace="mml40", name="", identifier=""):
        super().__init__(namespace=namespace, name=name, identifier=identifier)
        self.__displacement = None
        self.__min_displacement = None
        self.__max_displacement = None
        self.__displacements = []

    @property
    def displacement(self):
        return self.__displacement

    @displacement.setter
    def displacement(self, value):
        self.__displacement = value

    @property
    def maxDisplacement(self):
        return self.__max_displacement

    @maxDisplacement.setter
    def maxDisplacement(self, value):
        self.__max_displacement = value

    @property
    def minDisplacement(self):
        return self.__min_displacement

    @minDisplacement.setter
    def minDisplacement(self, value):
        self.__min_displacement = value

    @property
    def displacements(self):
        return self.__displacements

    @displacements.setter
    def displacements(self, value):
        self.__displacements = value

    def to_json(self):
        self.__json_out = super().to_json()
        if self.displacement is not None:
            self.__json_out["displacement"] = self.__displacement
        if self.__min_displacement is not None:
            self.__json_out["minDisplacement"] = self.__min_displacement
        if self.__max_displacement is not None:
            self.__json_out["maxDisplacement"] = self.__max_displacement
        if self.__displacements:
            self.__json_out["displacements"] = self.__displacements
        return self.__json_out



"""This module implements the role class according to Forest Modeling Language 4.0."""

import inspect
import sys
from abc import ABC

from ml.identifier import ID


class Role(ABC):
    """Each Role class manages a role which can be assigned to things."""

    def __init__(self, namespace, name="", identifier="", parent=None):
        """Constructs a new Role class object with the given name and
        identifier.

        :param name: Name of the feature
        :param identifier: Identifier of the feature
        """
        self.parent = parent
        self.__identifier = identifier
        self.__name = name
        if name == "":
            self.__name = self.__class__.__name__
        self.__class_name = f"{namespace}::{self.__class__.__name__}"
        self.__json_out = {}

    @property
    def class_name(self):
        """Adds the prefix ml40:: to the class name and returns it.

        :returns: class name of the object
        :rtype: str
        """
        return self.__class_name

    @property
    def name(self):
        """Returns the object's name.

        :returns: Name of the object
        :rtype: str

        """

        return self.__name

    @name.setter
    def name(self, value):
        """Sets __name to value.

        :param value: New name

        """

        self.__name = value

    @property
    def identifier(self):
        """Returns the role's local identifier.

        :returns: Identifier
        :rtype: str

        """

        return self.__identifier

    @identifier.setter
    def identifier(self, value):
        """Builds an ID object from value and assigns the resulting identifier
        to __identifier.

        :param value: Proposal of the identifier
        """

        self.__identifier = value

    def to_json(self):
        """Returns a JSON representation of this role."""

        self.__json_out = {"class": self.__class_name, "name": self.__name}
        if self.identifier:
            self.__json_out["identifier"] = self.identifier
        return self.__json_out

    def set_parent(self, value):
        self.parent = value



"""This module implements the class ForestPlanningEvaluation."""

from ml.ml40.features.functionalities.functionality import Functionality


class ForestPlanningEvaluation(Functionality):
    def __init__(self, namespace="fml40", name="", identifier="", parent=None):
        """Initializes the object.

        :param name: Object name
        :param identifier: Identifier

        """
        super().__init__(
            namespace=namespace, name=name, identifier=identifier, parent=parent
        )

    def evaluateInventoryData(self):
        pass



from ml.fml40.features.properties.values.documents.jobs.log_transportation_job import (
    LogTransportationJob,
)
from ml.fml40.features.properties.values.documents.reports.log_transportation_report import (
    LogTransportationReport,
)
from ml.ml40.features.functionalities.functionality import Functionality


class TransportsLogs(Functionality):
    def __init__(self, namespace="fml40", name="", identifier="", parent=None):
        super().__init__(
            namespace=namespace, name=name, identifier=identifier, parent=parent
        )

    def transportLogs(self, job: LogTransportationJob) -> LogTransportationReport:
        print("i am making Log transportation report for the job {}".format(job))
        return "the transportation report is huge"



from ml.ml40.features.properties.values.value import Value


class LiquidFillingLevel(Value):
    def __init__(self, namespace="ml40", name="", identifier="", parent=None):
        super().__init__(
            namespace=namespace, name=name, identifier=identifier, parent=parent
        )

        self.currentLevel = None
        self.maxLevel = None
        self.__json_out = dict()

    def to_json(self):
        self.__json_out = super().to_json()
        if self.currentLevel is not None:
            self.__json_out["currentLevel"] = self.currentLevel
        if self.maxLevel is not None:
            self.__json_out["maxLevel"] = self.maxLevel
        return self.__json_out



from ml.fml40.features.properties.values.abstract_inventory import \
    AbstractInventory
from ml.ml40.features.properties.values.value import Value


class InventoryData(Value):
    def __init__(self, namespace="fml40", name="", identifier="", parent=None):
        super().__init__(namespace=namespace, name=name, identifier=identifier, parent=parent)

        self.__data = []
        self.__json_out = dict()

    @property
    def data(self):
        return self.__data

    @data.setter
    def data(self, value):
        self.__data = value

    def to_json(self):
        self.__json_out = super().to_json()
        if self.data:
            self.__json_out["data"] = self.data
        return self.__json_out



from ml.ml40.features.properties.values.value import Value


class SwitchingStage(Value):
    def __init__(self, namespace="ml40", name="", identifier="", parent=None):
        super().__init__(
            namespace=namespace, name=name, identifier=identifier, parent=parent
        )

        self.__min_stage = None
        self.__current_stage = None
        self.__max_stage = None
        self.__json_out = dict()

    @property
    def minStage(self):
        return self.__min_stage

    @minStage.setter
    def minStage(self, value):
        self.__min_stage = value

    @property
    def currentStage(self):
        return self.__current_stage

    @currentStage.setter
    def currentStage(self, value):
        self.__current_stage = value

    @property
    def maxStage(self):
        return self.__max_stage

    @maxStage.setter
    def maxStage(self, value):
        self.__max_stage = value

    def to_json(self):
        self.__json_out = super().to_json()
        if self.minStage is not None:
            self.__json_out["minStage"] = self.minStage
        if self.maxStage is not None:
            self.__json_out["maxStage"] = self.maxStage
        if self.currentStage is not None:
            self.__json_out["currentStage"] = self.currentStage
        return self.__json_out



from ml.ml40.features.properties.values.value import Value


class LastServiceCheck(Value):
    def __init__(self, namespace="ml40", name="", identifier="", parent=None):
        super().__init__(
            namespace=namespace, name=name, identifier=identifier, parent=parent
        )

        self.__timestamp = None
        self.__json_out = dict()

    @property
    def timestamp(self):
        return self.__timestamp

    @timestamp.setter
    def timestamp(self, value):
        self.__timestamp = value

    def to_json(self):
        self.__json_out = super().to_json()
        if self.__timestamp is not None:
            self.__json_out["weight"] = self.__timestamp
        return self.__json_out



from ml.ml40.features.properties.values.value import Value


class StereoCameraConfig(Value):
    def __init__(self, namespace="ml40", name="", identifier=""):
        super().__init__(
            namespace=namespace,
            name=name,
            identifier=identifier)

        self.__frame_rate = None
        self.__disparity_format = None

    @property
    def frameRate(self):
        return self.__frame_rate

    @frameRate.setter
    def frameRate(self, value):
        self.__frame_rate = value

    @property
    def disparityFormat(self):
        return self.__disparity_format

    @disparityFormat.setter
    def disparityFormat(self, value):
        self.__disparity_format = value

    def to_json(self):
        self.__json_out = super().to_json()
        if self.__frame_rate is not None:
            self.__json_out["frameRate"] = self.__frame_rate
        if self.__disparity_format is not None:
            self.__json_out["disparityFormat"] = self.__disparity_format
        return self.__json_out


from ml.role import Role


class Service(Role):
    def __init__(self, namespace="ml40", name="", identifier="", parent=None):
        super(Service, self).__init__(
            namespace=namespace, name=name, identifier=identifier, parent=parent
        )



from ml.ml40.features.functionalities.functionality import Functionality
from ml.ml40.features.properties.values.force import Force


class ProvidesForceData(Functionality):
    def __init__(self, namespace="mml40", name="", identifier="", parent=None):
        """Initializes the object.

        :param name: Object name
        :param identifier: Identifier

        """
        super().__init__(
            namespace=namespace, name=name, identifier=identifier, parent=parent
        )

    def getMaxForce(self) -> Force:
        pass

    def getMinForce(self) -> Force:
        pass

    def getForceData(self, time) -> Force:
        pass

    def getForceDataSeries(self, startTime, endTime) -> [Force]:
        pass



"""This modules provides the ID class which can be used to build S3I
specific identifiers."""

import uuid


class ID:
    """Define primitive variable type ID."""

    def __init__(self, identifier=""):
        """Constructs an ID object with the given identifier.

        :param identifier: Proposal of the identifier

        """

        # TODO validate identifier
        is_valid = self.is_valid_uuid(identifier.replace("s3i:", ""))
        if identifier != "" and is_valid:
            self.__identifier = identifier
        else:
            self.__identifier = "s3i:{}".format(uuid.uuid4())

    @property
    def identifier(self):
        """Returns the identifier.

        :returns: Identifier
        :rtype: str

        """

        return self.__identifier

    @staticmethod
    def is_valid_uuid(uuid_to_test, version=4):
        """Returns true if uuid_to_test is a valid uuid, otherwise return
        false.

        :param uuid_to_test: Uuid to be tested
        :param version: Uuid version
        :returns: Test result
        :rtype: bool

        """

        try:
            uuid_obj = uuid.UUID(uuid_to_test, version=version)
        except ValueError:
            return False
        return str(uuid_obj) == uuid_to_test



from ml.ml40.features.functionalities.functionality import Functionality
from ml.mml40.features.properties.values.Displacement import Displacement


class ProvidesDisplacementData(Functionality):
    def __init__(self, namespace="mml40", name="", identifier="", parent=None):
        """Initializes the object.

        :param name: Object name
        :param identifier: Identifier

        """
        super().__init__(namespace=namespace, name=name, identifier=identifier, parent=parent)

    def getMaxDisplacement(self) -> Displacement:
        pass

    def getMinDisplacement(self) -> Displacement:
        pass

    def getDisplacementData(self, time) -> [Displacement]:
        pass

    def getDisplacementDataSeries(self, startTime, endTime) -> [Displacement]:
        pass



"""This module implements the class GeneratesAfforestationSuggestions."""

from ml.identifier import ID
from ml.ml40.features.functionalities.functionality import Functionality


class GeneratesFellingSuggestions(Functionality):
    """This functionality can generate suggestions regarding how to fell a
    tree.
    """

    def __init__(self, namespace="fml40", name="", identifier="", parent=None):
        """Initializes the object.

        :param name: Object name
        :param identifier: Identifier

        """
        super().__init__(
            namespace=namespace, name=name, identifier=identifier, parent=parent
        )

    def generateFellingSuggestion(self, tree_Id: ID):
        """Returns a suggestion regarding how to fell a tree.

        :param: tree_Id: Identifier of a tree
        """
        pass



from ml.dt_factory import create_thing, build, build_feature, build_role, build_sub_features, build_sub_thing
from ml.tools import make_thing_config, load_config, make_feature_config, find_broker_endpoint
from ml.app_logger import setup_logger, APP_LOGGER
from ml.entry import Entry
from ml.parameters import Parameters
from ml.callback import CallbackManager



from ml.ml40.features.properties.values.value import Value


class GeometryProperties(Value):
    def __init__(self, namespace="mml40", name="", identifier=""):
        super().__init__(namespace=namespace, name=name, identifier=identifier)
        self.__geometry_type = None
        self.__length = None
        self.__width = None
        self.__height = None
        self.__thickness = None
        self.__json_out = dict()

    @property
    def geometryType(self):
        return self.__geometry_type

    @geometryType.setter
    def geometryType(self, value):
        self.__geometry_type = value

    @property
    def thickness(self):
        return self.__thickness

    @thickness.setter
    def thickness(self, value):
        self.__thickness = value

    @property
    def length(self):
        return self.__length

    @length.setter
    def length(self, value):
        self.__length = value

    @property
    def width(self):
        return self.__width

    @width.setter
    def width(self, value):
        self.__width = value

    @property
    def height(self):
        return self.__height

    @height.setter
    def height(self, value):
        self.__height = value

    def to_json(self):
        self.__json_out = super().to_json()
        if self.__geometry_type is not None:
            self.__json_out["geometryType"] = self.__geometry_type

        if self.__length is not None:
            self.__json_out["length"] = self.__length

        if self.__width is not None:
            self.__json_out["width"] = self.__width

        if self.__height is not None:
            self.__json_out["height"] = self.__height

        if (
            self.__thickness is not None
            and self.__geometry_type == "rectangular_hollow_profile"
        ):
            self.__json_out["thickness"] = self.__thickness
        else:
            self.__json_out["thickness"] = ""

        return self.__json_out



from ml.ml40.features.properties.values.value import Value


class Lift(Value):
    def __init__(self, namespace="ml40", name="", identifier="", parent=None):
        super().__init__(
            namespace=namespace, name=name, identifier=identifier, parent=parent
        )

        self.__current_lift = None
        self.__max_lift = None
        self.__min_lift = None
        self.__json_out = dict()

    @property
    def currentLift(self):
        return self.__current_lift

    @currentLift.setter
    def currentLift(self, value):
        self.__current_lift = value

    @property
    def maxLift(self):
        return self.__max_lift

    @maxLift.setter
    def maxLift(self, value):
        self.__max_lift = value

    @property
    def minLift(self):
        return self.__min_lift

    @minLift.setter
    def minLift(self, value):
        self.__min_lift = value

    def to_json(self):
        self.__json_out = super().to_json()
        if self.__current_lift is not None:
            self.__json_out["currentLift"] = self.__current_lift
        if self.__max_lift is not None:
            self.__json_out["maxLift"] = self.__max_lift
        if self.__min_lift is not None:
            self.__json_out["minLift"] = self.__min_lift

        return self.__json_out



from ml.fml40.features.properties.values.documents.reports.passability_report import \
    PassabilityReport
from ml.ml40.features.functionalities.functionality import Functionality


class ProvidesMapData(Functionality):
    def __init__(self,namespace="ml40", name="", identifier="", parent=None):
        super().__init__(
            namespace=namespace,
            name=name,
            identifier=identifier,
            parent=parent)

    def getMapData(self, map: int) -> PassabilityReport:
        pass



from ml.ml40.features.properties.values.value import Value


class Dimensions(Value):
    def __init__(self, namespace="ml40", name="", identifier="", parent=None):
        super().__init__(
            namespace=namespace, name=name, identifier=identifier, parent=parent
        )

        self.height = None
        self.length = None
        self.weight = None
        self.width = None
        self.__json_out = dict()

    def to_json(self):
        self.__json_out = super().to_json()
        if self.height is not None:
            self.__json_out["height"] = self.height
        if self.length is not None:
            self.__json_out["length"] = self.length
        if self.weight is not None:
            self.__json_out["weight"] = self.weight
        if self.width is not None:
            self.__json_out["width"] = self.width

        return self.__json_out



import asyncio
import sys

import s3i
from s3i.identity_provider import IdentityProvider, TokenType
from s3i.directory import Directory
from s3i.repository import Repository
# from s3i.broker import BrokerREST, Broker
from s3i.event_system import EventManager
from ml.s3i_tools import Broker, BrokerREST
from s3i.broker_message import GetValueReply, ServiceReply, SetValueReply, CustomEventReply, GetValueRequest, \
    ServiceRequest, SetValueRequest
from s3i.exception import raise_error_from_s3ib_msg, S3IBMessageError, S3IDittoError, S3IIdentityProviderError
from ml.app_logger import APP_LOGGER
from ml.tools import find_broker_endpoint
from ml.callback import CallbackManager
from ml.parameters import Parameters
from ast import literal_eval
import json
import uuid
import copy


class Thing:
    _ON_THING_START_OK = "_on_thing_start_ok"
    _ON_IDP_START_OK = "_on_idp_start_ok"
    _ON_DIRECTORY_START_OK = "_on_directory_start_ok"
    _ON_REPOSITORY_START_OK = "_on_repository_start_ok"
    _ON_BROKER_START_OK = "_on_broker_start_ok"

    def __init__(
            self,
            entry,
            oauth2_secret,
            grant_type,
            username=None,
            password=None,
            is_repository=False,
            is_broker=False,
            is_broker_rest=False,
            is_event_manager=False,
            parameters=Parameters(),
            loop=asyncio.get_event_loop()

    ):
        self.__entry = entry
        self.__secret = oauth2_secret
        self.__grant_type = grant_type
        self.__username = username
        self.__password = password
        self.__token = None
        self.__is_repository = is_repository
        self.__is_broker = is_broker
        self.__is_broker_rest = is_broker_rest
        self.__is_event_manager = is_event_manager
        self.loop = loop
        self.__resGetValue = []

        self.callbacks = CallbackManager()
        self.parameters = parameters

    @property
    def entry(self):
        return self.__entry

    @property
    def identifier(self):
        return self.__entry.identifier

    def run_forever(self):
        try:
            self.__setup_thing_json_sync()
            self.__connect_to_s3i()
            self.loop.run_forever()

        except KeyboardInterrupt:
            APP_LOGGER.info("[S3I]: Disconnect from S3I")
            self.loop.close()

    def add_on_thing_start_ok_callback(self, callback_func, one_shot, is_async, *args, **kwargs):
        self.callbacks.add(
            self._ON_THING_START_OK,
            callback_func,
            one_shot,
            is_async,
            *args,
            **kwargs
        )

    def add_on_idp_start_ok_callback(self, callback_func, one_shot, is_async, *args, **kwargs):
        self.callbacks.add(
            self._ON_IDP_START_OK,
            callback_func,
            one_shot,
            is_async,
            *args,
            **kwargs
        )

    def add_on_directory_start_ok_callback(self, callback_func, one_shot, is_async, *args, **kwargs):
        self.callbacks.add(
            self._ON_DIRECTORY_START_OK,
            callback_func,
            one_shot,
            is_async,
            *args,
            **kwargs
        )

    def add_on_repository_start_ok_callback(self, callback_func, one_shot, is_async, *args, **kwargs):
        self.callbacks.add(
            self._ON_REPOSITORY_START_OK,
            callback_func,
            one_shot,
            is_async,
            *args,
            **kwargs
        )

    def add_on_broker_start_ok_callback(self, callback_func, one_shot, is_async, *args, **kwargs):
        def _add_on_channel_open_callback(_thing, _callback, _one_shot, *_args, **_kwargs):
            _thing.broker.add_on_channel_open_callback(
                callback_func,
                one_shot,
                *args,
                **kwargs
            )
        self.callbacks.add(
            self._ON_BROKER_START_OK,
            _add_on_channel_open_callback,
            one_shot,
            is_async, # the following parameters are *args
            self,
            callback_func,
            one_shot,
            *args,
            **kwargs
        )

    def add_ml40_implementation(self, implementation_object, ml40_feature, *args, **kwargs):
        def _add_ml40_implementation(_thing, _implementation_object, _ml40_feature, *_args, **_kwargs):
            feature = _thing.entry.features.get(_ml40_feature, None)
            if feature is None:
                APP_LOGGER.critical(
                    "Functionality %s is not one of the build-in functionalities" % _ml40_feature
                )
            else:
                APP_LOGGER.info("Implementation object is added into the functionality %s" % _ml40_feature)
                impl_ins = _implementation_object(*_args, **_kwargs)
                impl_ins.class_name = _ml40_feature
                _thing.entry.features[_ml40_feature] = impl_ins

        self.add_on_thing_start_ok_callback(
            _add_ml40_implementation,
            True,
            False,
            self,
            implementation_object,
            ml40_feature,
            *args,
            **kwargs
        )

    def __connect_to_s3i(self):
        self.__setup_identity_provider()
        self.__setup_directory()
        if self.__is_repository:
            self.__setup_repository()
        if self.__is_broker:
            self.__setup__broker()
        if self.__is_event_manager:
            self.__setup_event_system()

    def __setup_identity_provider(self):
        APP_LOGGER.info("[S3I]: Connect to IdentityProvider")
        self.idp = IdentityProvider(
            grant_type=self.__grant_type,
            identity_provider_url=self.parameters.idp_url,
            realm=self.parameters.idp_realm,
            client_id=self.identifier,
            client_secret=self.__secret,
            username=self.__username,
            password=self.__password
        )
        try:
            self.__token = self.idp.get_token(TokenType.ACCESS_TOKEN)
        except S3IIdentityProviderError as err:
            APP_LOGGER.error("[S3I]: {}".format(err))
        else:
            APP_LOGGER.info("[S3I]: Access Token granted")
        self.loop.call_later(self.__get_remaining_time_to_refresh(),
                             self.__refresh_token_recur)
        self.callbacks.process(
            self._ON_IDP_START_OK,
            self.loop
        )

    def __refresh_token_recur(self):
        APP_LOGGER.info("[S3I]: Get refreshed access token from Identity Provider")
        self.idp._refresh_token(self.idp.get_token(TokenType.REFRESH_TOKEN))
        # Keycloak returns sometimes a invalid access token, even through it was gained via valid refresh token
        # Bug
        if self.idp._token_bundle["expires_in"] < 60:
            self.idp._authenticate(scope="openid")

        self.__token = self.idp._token_bundle["access_token"]

        # refresh s3i object
        APP_LOGGER.info("[S3I]: Refresh token to Directory")
        self.dir.pass_refreshed_token(self.__token)
        if self.__is_repository:
            APP_LOGGER.info("[S3I]: Refresh token to Repository")
            self.repo.pass_refreshed_token(self.__token)
        if self.__is_broker:
            self.broker.reconnect_token_expired(self.__token)
        self.loop.call_later(self.__get_remaining_time_to_refresh(),
                             self.__refresh_token_recur)

    def __get_remaining_time_to_refresh(self):
        remaining_time = self.idp._time_until_token_valid()
        safety_margin = 5
        return remaining_time - safety_margin

    def __setup_thing_json_sync(self):
        APP_LOGGER.info("[S3I]: Start the thing")
        self.__recursively_update_dt_json(frequency=self.parameters.thing_sync_freq)
        self.callbacks.process(
            prefix=self._ON_THING_START_OK,
            loop=self.loop
        )

    def __setup_directory(self):
        APP_LOGGER.info("[S3I]: Connect to Directory")
        self.dir = Directory(
            s3i_dir_url=self.parameters.dir_url,
            token=self.__token
        )
        self.__recursively_update_directory(frequency=self.parameters.dir_sync_freq)
        self.callbacks.process(
            self._ON_DIRECTORY_START_OK,
            self.loop
        )

    def __setup__broker(self):
        # TODO Bug: by network disconnect the connection to broker keeps still connected
        # TODO allow having more channels + consumers
        # TODO send via BrokerREST()
        endpoint = find_broker_endpoint(self.dir, self.identifier)
        APP_LOGGER.info("[S3I]: Connecting to Broker")
        if self.__is_broker_rest:
            self.broker = BrokerREST(token=self.__token)
        else:
            self.broker = Broker(
                token=self.__token,
                endpoint=endpoint,
                callback=self.on_broker_message_callback,
                loop=self.loop
            )
            self.broker.connect()
        self.callbacks.process(
            self._ON_BROKER_START_OK,
            self.loop
            )

    def __setup_repository(self):
        APP_LOGGER.info("[S3I]: Connect to Repository")
        self.repo = Repository(
            s3i_repo_url=self.parameters.repo_url,
            token=self.__token
        )
        self.__recursively_update_repository(frequency=self.parameters.repo_sync_freq)
        self.callbacks.process(
            self._ON_REPOSITORY_START_OK,
            self.loop
        )

    def __setup_event_system(self):
        # TODO implement event system
        self.event_manager = EventManager(
            json_entry=self.entry.dt_json
        )
        """
        self.event_manager = EventManager(
        )
        """

    def __recursively_update_directory(self, frequency):
        try:
            self.dir.updateThingIDBased(thingID=self.identifier, payload=self.__entry.refresh_directory_entry(
                current_dir_json=self.dir.queryThingIDBased(self.identifier)
            ))
        except S3IDittoError as err:
            APP_LOGGER.error("[S3I]: {}".format(err))
        if frequency != 0:
            self.loop.call_later(
                1 / frequency,
                self.__recursively_update_directory,
                frequency
            )

    def __recursively_update_repository(self, frequency):
        try:
            self.repo.updateThingIDBased(thingID=self.identifier, payload=self.__entry.refresh_repository_entry())
        except S3IDittoError as err:
            APP_LOGGER.error("[S3I]: {}".format(err))
        if frequency != 0:
            self.loop.call_later(
                1 / frequency,
                self.__recursively_update_repository,
                frequency
            )

    def __recursively_update_dt_json(self, frequency):
        self.__entry.refresh_dt_json()
        if frequency != 0:
            self.loop.call_later(
                1 / frequency,
                self.__recursively_update_dt_json,
                frequency
            )

    def on_broker_message_callback(self, ch, method, properties, body):
        """Parses body (content of a S3I-B message) and delegates the
        processing of the message to a separate method. The method is
        selected according to the message's type.

        :param body: S3I-B message

        """
        try:
            body = literal_eval(body.decode('utf-8'))
        except ValueError:
            body = json.loads(body)
        # Add S3I-B Message validate
        try:
            body = raise_error_from_s3ib_msg(body, S3IBMessageError)
        except S3IBMessageError as e:
            APP_LOGGER.error("[S3I]: {}".format(e))
        else:
            message_type = body.get("messageType")

            __log = "[S3I]: Received a S3I-B {}: {}".format(
                message_type, json.dumps(body, indent=2)
            )
            APP_LOGGER.info(__log)

            if message_type == "userMessage":
                self.on_user_message(body)
            elif message_type == "serviceRequest":
                self.loop.create_task(self.on_service_request(body))
            elif message_type == "getValueRequest":
                self.on_get_value_request(body)
            elif message_type == "setValueRequest":
                self.on_set_value_request(body)
            elif message_type == "eventSubscriptionRequest":
                self.on_event_subscription_request(body)
            else:
                pass

    def on_user_message(self, msg):
        """Handles incoming S³I-B UserMessages.

        :param msg: S³I-B UserMessages

        """
        pass

    def on_get_value_request(self, msg):
        """Handles incoming GetValueRequest message. Looks up the value specified in msg and
        sends a GetValueReply message back to the sender.

        :param msg: GetValueRequest

        """
        req = GetValueRequest(base_msg=msg)
        request_sender = req.base_msg.get("sender")
        request_msg_id = req.base_msg.get("identifier")
        request_sender_endpoint = req.base_msg.get("replyToEndpoint")
        attribute_path = req.base_msg.get("attributePath")
        reply_msg_uuid = "s3i:" + str(uuid.uuid4())

        try:
            __log = "[S3I]: Search the given attribute path: {}".format(attribute_path)
            APP_LOGGER.info(__log)
            value = self._uriToData(attribute_path)
        except KeyError:
            value = "Invalid attribute path"
            __log = "[S3I]: " + value
            APP_LOGGER.critical(__log)

        get_value_reply = GetValueReply()
        get_value_reply.fillGetValueReply(
            sender=self.identifier,
            receivers=[request_sender],
            message_id=reply_msg_uuid,
            replying_to_msg=request_msg_id,
            value=value
        )

        res = self.broker.send(
            endpoint=request_sender_endpoint,
            msg=get_value_reply.base_msg
        )
        if res:
            APP_LOGGER.info("[S3I]: Sending getValueReply back")
        else:
            APP_LOGGER.error("[S3I]: Sending getValueReply failed")

    def _uriToData(self, uri):
        """Returns a copy of the value found at uri.

        :param uri: Path to value
        :rtype: Feature

        """

        if uri == "":
            return self.__entry.dt_json
        else:
            uri_list = uri.split("/")
            if uri_list[0] == "features":
                try:
                    return self.__entry.dt_json[uri]
                except KeyError:
                    return "Invalid attribute path"

            try:
                self._getValue(self.__entry.dt_json, uri_list)
            except Exception:
                return "Invalid attribute path"
            if self.__resGetValue.__len__() == 0:
                return "Invalid attribute path"
            response = copy.deepcopy(self.__resGetValue)
            self.__resGetValue.clear()
            if response.__len__() == 1:
                return response[0]
            else:
                return response

    def _getValue(self, source, uri_list):
        """Searches for the value specified by uri_list in source and stores
        the result in __resGetValue.

        :param source: Object that is scanned
        :param uri_list: List containing path

        """

        # ??? What if the uri points to a Value object?
        # Shouldn't it be serialized?!
        value = source[uri_list[0]]
        if uri_list.__len__() == 1:
            # if is ditto-feature
            if isinstance(value, str):
                try:
                    stringValue_split = value.split(":")
                    if stringValue_split[0] == "ditto-feature":
                        value = self.__entry.dt_json["features"][stringValue_split[1]][
                            "properties"
                        ][uri_list[0]]
                except Exception:
                    pass
            self.__resGetValue.append(value)
            return
        if isinstance(value, dict):
            # ??? uri_list.pop(0) better?!
            del uri_list[0]
            self._getValue(value, uri_list)
        if isinstance(value, list):
            if isinstance(value[0], (str, int, float, bool, list)):
                return value
            if isinstance(value[0], dict):
                for item in value:
                    if item["class"] == "ml40::Thing":
                        for i in item["roles"]:
                            if self._findValue(i, uri_list[1]):
                                uri_list_1 = copy.deepcopy(uri_list)
                                del uri_list_1[:2]
                                self._getValue(item, uri_list_1)
                        _f = self._findValue({"identifier": item.get("identifier")}, uri_list[1]) or \
                             self._findValue({"name": item.get("name")}, uri_list[1])
                        if _f:
                            uri_list_1 = copy.deepcopy(uri_list)
                            del uri_list_1[:2]
                            self._getValue(item, uri_list_1)
                    else:
                        if self._findValue(item, uri_list[1]):
                            uri_list_1 = copy.deepcopy(uri_list)
                            del uri_list_1[:2]
                            if not uri_list_1:
                                self.__resGetValue.append(item)
                                return
                            else:
                                self._getValue(item, uri_list_1)
        if isinstance(value, (str, int, float, bool)):
            # if is ditto-feature
            if isinstance(value, str):
                try:
                    stringValue_split = value.split(":")
                    if stringValue_split[0] == "ditto-feature":
                        value = self.__entry.dt_json["features"][stringValue_split[1]][
                            "properties"
                        ][uri_list[0]]
                except Exception:
                    pass
            self.__resGetValue.append(value)

    def _findValue(self, dic, value):
        """Returns true if value has been found in json, otherwise returns false.

        :param dic: dictionary
        :param value:
        :returns:
        :rtype:

        """

        for val in dic.values():
            if val == value:
                return True
        return False

    async def on_service_request(self, body_json):
        """Handles S³I-B ServiceRequests. Executes the method of the
        functionality specified in serviceType and send a ServiceReply
        back to the sender.

        :param body_json: ServiceRequest

        """
        req = ServiceRequest(base_msg=body_json)
        service_type = req.base_msg.get("serviceType")
        parameters = req.base_msg.get("parameters")
        request_sender = req.base_msg.get("sender")
        request_id = req.base_msg.get("identifier")

        service_reply = ServiceReply()
        service_functionality = service_type.split('/')[0]
        service_functionality_obj = self.entry.features.get(service_functionality)
        if service_functionality_obj is None:
            APP_LOGGER.critical(
                "[S3I]: Functionality %s is not one of the built-in functionalities in %s!"
                % (service_functionality, self.entry.name)
            )
            service_reply.fillServiceReply(
                sender=self.identifier,
                receivers=[request_sender],
                service_type=service_type,
                results={"error": "invalid functionalities (serviceType) {}".format(service_functionality)},
                replying_to_msg=request_id,
                message_id="s3i:{}".format(uuid.uuid4())
            )
        else:
            # TODO: Call right functionality.
            try:
                method = getattr(service_functionality_obj, service_type.split('/')[1])
            except AttributeError:
                APP_LOGGER.critical(
                    "[S3I]: Method %s is not one of the built-in functionalities in %s!" % (
                        service_type.split('/')[1], self.entry.name)
                )
                service_reply.fillServiceReply(
                    sender=self.identifier,
                    receivers=[request_sender],
                    service_type=service_type,
                    replying_to_msg=request_id,
                    message_id="s3i:{}".format(uuid.uuid4()),
                    results={"error": "invalid method {}".format(service_type.split('/')[1])},
                )
            except IndexError:
                APP_LOGGER.critical(
                    "[S3I]: ServiceType consists of functionality and method name."
                )
                service_reply.fillServiceReply(
                    sender=self.identifier,
                    receivers=[request_sender],
                    service_type=service_type,
                    replying_to_msg=request_id,
                    results={"error": "method missing"},
                    message_id="s3i:{}".format(uuid.uuid4())
                )
            else:
                __log = "[S3I]: Execute the function {0} of the class {1}".format(service_type.split('/')[1],
                                                                                  service_type.split('/')[0])
                APP_LOGGER.info(__log)
                try:
                    result = await method(**parameters)
                except TypeError:
                    APP_LOGGER.critical("[S3I]: Invalid function arguments")
                    service_reply.fillServiceReply(
                        sender=self.identifier,
                        receivers=[request_sender],
                        service_type=service_type,
                        replying_to_msg=request_id,
                        results={"error": "invalid function arguments (parameters)"},
                        message_id="s3i:{}".format(uuid.uuid4())
                    )
                else:
                    if isinstance(result, bool):
                        result = {"ok": result}
                    elif result is None:
                        result = "None"
                    service_reply.fillServiceReply(
                        sender=self.identifier,
                        receivers=[request_sender],
                        service_type=service_type,
                        replying_to_msg=request_id,
                        results=result,
                        message_id="s3i:{}".format(uuid.uuid4())
                    )
        if sys.getsizeof(service_reply.base_msg["results"]) > self.parameters.broker_msg_compress_threshold:
            service_reply.compress(msg_json=service_reply.base_msg, level=6)
            res = self.broker.send(
                endpoint=body_json.get("replyToEndpoint", None),
                msg=service_reply.gzip_msg
            )
        else:
            res = self.broker.send(
                endpoint=body_json.get("replyToEndpoint", None),
                msg=service_reply.base_msg)

        if res:
            APP_LOGGER.info("[S3I]: Sending serviceReply back")
        else:
            APP_LOGGER.error("[S3I]: Sending serviceReply failed")

    def on_set_value_request(self, msg):
        """Handles incoming S³I-B SetValueRequest. Prints the content of msg to stdout.

        :param msg: GetValueReply

        """
        set_value_reply = SetValueReply()

        req = SetValueRequest(base_msg=msg)

        request_sender = req.base_msg.get("sender")
        request_msg_id = req.base_msg.get("identifier")
        request_sender_endpoint = req.base_msg.get("replyToEndpoint")
        attribute_path = req.base_msg.get("attributePath")
        new_value = req.base_msg.get("newValue")
        reply_msg_uuid = "s3i:" + str(uuid.uuid4())

        try:
            __log = "[S3I]: Search for the given attribute path: {}".format(attribute_path)
            APP_LOGGER.info(__log)
            old_value = self._uriToData(attribute_path)
            ins = self._uriToIns(attribute_path)
            APP_LOGGER.info("[S3I]: Change value from {} to {}".format(old_value, new_value))
            result = self._set_value_req(ins, new_value, attribute_path)

        except Exception:
            __log = "[S3I]: Invalid attribute path"
            APP_LOGGER.critical(__log)
            result = False

        set_value_reply.fillSetValueReply(
            sender=self.identifier,
            receivers=[request_sender],
            ok=result,
            replying_to_msg=request_msg_id,
            message_id=reply_msg_uuid
        )
        res = self.broker.send(
            endpoint=request_sender_endpoint,
            msg=set_value_reply.base_msg
        )
        if res:
            APP_LOGGER.info("[S3I]: Sending setValueReply back")
        else:
            APP_LOGGER.error("[S3I]: Sending setValueReply failed")

    def _set_value_req(self, ins, new_value, attribute_path):
        if not isinstance(new_value, dict):
            attr_list = attribute_path.split("/")
            if attr_list.__len__() <= 2:
                APP_LOGGER.info("Not allowed to set attribute {}".format(attribute_path))
                return False
            else:
                if hasattr(ins, attr_list[attr_list.__len__() - 1]):
                    setattr(ins, attr_list[attr_list.__len__() - 1], new_value)
                    return True
                APP_LOGGER.info("{} is not one of the attributes".format(attr_list[attr_list.__len__() - 1]))
                return False
        else:
            for key in new_value.keys():
                if hasattr(ins, key):
                    setattr(ins, key, new_value[key])
                else:
                    APP_LOGGER.info("{} is not one of the attributes".format(key))
                    return False
            return True

    def _uriToIns(self, uri):
        if not uri:
            return None
        uri_list = uri.split("/")
        uri_list.pop(0)  # delete first element "attributes"
        return self._getInstance(self, uri_list)

    def _getInstance(self, source_obj, uri_list):
        if uri_list.__len__() == 0 or uri_list.__len__() == 1:
            ### the original uri was "attributes/features"
            return source_obj

        if "ml40" in uri_list[0]:
            _uri = uri_list[0]
            uri_list.pop(0)
            return self._getInstance(source_obj.features[_uri], uri_list)

        elif uri_list[0] == "features":
            uri_list.pop(0)
            return self._getInstance(source_obj, uri_list)

        elif uri_list[0] == "targets":
            uri_list.pop(0)
            for key in source_obj.targets.keys():
                subthing_dict = source_obj.targets[key].to_subthing_json()
                if subthing_dict.get("name", "") == uri_list[0] or subthing_dict.get("identifier", "") == uri_list[0] \
                        or subthing_dict.get("class", "") == uri_list[0]:
                    uri_list.pop(0)
                    return self._getInstance(source_obj.targets[key], uri_list)

        elif uri_list[0] == "subFeatures":
            uri_list.pop(0)
            for key in source_obj.subFeatures.keys():
                subfeature_dict = source_obj.subFeatures[key].to_json()
                if subfeature_dict.get("name", "") == uri_list[0] or subfeature_dict.get("identifier", "") == uri_list[
                    0] \
                        or subfeature_dict.get("class", "") == uri_list[0]:
                    uri_list.pop(0)
                    return self._getInstance(source_obj.subFeatures[key], uri_list)

    def on_event_subscription_request(self, msg):
        subscription_status, topic = self.event_manager.add_custom_event(
            rql_expression=msg.get("filter"),
            attribute_paths=msg.get("attributePaths")
        )
        __log = "[S3I][Broker]: Validation of RQL syntax: {}".format(subscription_status)
        APP_LOGGER.info(__log)
        event_sub_reply = CustomEventReply()
        event_sub_reply.fillCustomEventReply(
            sender=self.identifier,
            receivers=[msg.get("sender")],
            topic=topic,
            replying_to_msg=msg.get("identifier"),
            message_id="s3i:" + str(uuid.uuid4()),
            status="ok" if subscription_status else "invalid request"
        )
        res = self.broker.send(
            msg.get("replyToEndpoint"),
            msg=event_sub_reply.base_msg
        )
        if res:
            APP_LOGGER.info("[S3I]: Sending customEventReply back")
        else:
            APP_LOGGER.error("[S3I]: Sending customEventReply failed")



from ml.fml40.features.properties.values.documents.jobs.felling_job import FellingJob
from ml.ml40.features.functionalities.functionality import Functionality


class SupportsFelling(Functionality):
    def __init__(self, namespace="fml40", name="", identifier="", parent=None):
        super().__init__(
            namespace=namespace, name=name, identifier=identifier, parent=parent
        )

    def supportFelling(self, job: FellingJob, suggestion):
        pass



from ml.ml40.roles.dts.machines.machine import Machine


class ForestMachine(Machine):
    def __init__(self, namespace="fml40", name="", identifier="", parent=None):
        super().__init__(namespace=namespace, name=name, identifier=identifier, parent=parent)



from ml.ml40.features.functionalities.functionality import Functionality
from ml.ml40.features.properties.values.documents.reports.report import Report


class AcceptsReports(Functionality):
    def __init__(self,namespace="ml40", name="", identifier="", parent=None):
        super().__init__(
            namespace=namespace,
            name=name,
            identifier=identifier,
            parent=parent)

    def acceptReport(self, report: Report):
        pass



"""This module implements the class GeneratesAfforestationSuggestions."""

from ml.fml40.features.properties.values.documents.reports.afforestation_suggestion import (
    AfforestationSuggestion,
)
from ml.ml40.features.functionalities.functionality import Functionality


class GeneratesAfforestationSuggestions(Functionality):
    """This functionality can generate suggestions for afforestation."""

    def __init__(self, namespace="fml40", name="", identifier="", parent=None):
        """Initializes the object.

        :param name: Object name
        :param identifier: Identifier

        """
        super().__init__(
            namespace=namespace, name=name, identifier=identifier, parent=parent
        )

    def generateAfforestationSuggestion(self) -> AfforestationSuggestion:
        """Returns a suggestion for afforestation.

        :returns: Suggestion for afforestation
        :rtype: AfforestationSuggestion
        """
        pass



from ml.ml40.features.properties.values.value import Value


class Force(Value):
    def __init__(self, namespace="ml40", name="", identifier="", parent=None):
        super().__init__(
            namespace=namespace, name=name, identifier=identifier, parent=parent
        )
        self.__force = None
        self.__min_force = None
        self.__max_force = None

    @property
    def force(self):
        return self.__force

    @force.setter
    def force(self, value):
        self.__force = value

    @property
    def minForce(self):
        return self.__min_force

    @minForce.setter
    def minForce(self, value):
        self.__min_force = value

    @property
    def maxForce(self):
        return self.__max_force

    @maxForce.setter
    def maxForce(self, value):
        self.__max_force = value

    def to_json(self):
        self.__json_out = super().to_json()
        if self.__force is not None:
            self.__json_out["force"] = self.__force

        if self.__min_force is not None:
            self.__json_out["minForce"] = self.__min_force

        if self.__max_force is not None:
            self.__json_out["maxForce"] = self.__max_force

        return self.__json_out



from ml.fml40.features.properties.values.documents.reports.passability_report import \
    PassabilityReport
from ml.ml40.features.functionalities.functionality import Functionality


class ProvidesMoisturePrediction(Functionality):
    def __init__(self, namespace="fml40", name="", identifier="", parent=None):
        super().__init__(namespace=namespace, name=name, identifier=identifier, parent=parent)

    def predict(self):
        pass

    def predictMoistures(self, location, moisture):
        pass



from ml.ml40.features.properties.values.documents.jobs.job import Job


class GenericJob(Job):
    def __init__(self, namespace="ml40", name="", identifier="", parent=None):
        super().__init__(
            namespace=namespace, name=name, identifier=identifier, parent=parent
        )
        self.__content = None
        self.__json_out = dict()

    @property
    def content(self):
        return self.__content

    @content.setter
    def content(self, value):
        self.__content = value

    def to_json(self):
        self.__json_out = super().to_json()
        if self.content is not None:
            self.__json_out["content"] = self.content

        return self.__json_out



"""This module implements the Feature class according to the Forest
Modeling Language 4.0."""

import inspect
import os
import sys
from abc import ABC

from ml.identifier import ID
from ml.thing import Thing


class Feature(ABC):
    """The Feature class represents the base class for all functionalities
    and values."""

    def __init__(self, namespace, name="", identifier="", parent=None):
        """Constructs a new Feature class object with the given name and
        identifier.

        :param name: Name of the feature
        :param identifier: Identifier of the feature
        """
        self.parent = parent
        self.namespace = namespace
        self.__name = name
        self.__class_name = f"{self.namespace}::{self.__class__.__name__}"

        self.__identifier = identifier
        self.__subFeatures = dict()
        self.__json_out = dict()

    @property
    def class_name(self):
        """Adds the prefix ml40:: to the class name and returns it.

        :returns: Name of this object's class
        :rtype: str

        """

        return self.__class_name

    @class_name.setter
    def class_name(self, value):
        """Sets __class_name to value.

        :param value: New class name
        :type value: str

        """
        # !!! What about the namespace?
        # ??? Do we really need to have this function?
        # Can't this be done implicitly?
        self.__class_name = value

    @property
    def name(self):
        """Returns the object's name.

        :returns: Name of the object
        :rtype: str

        """

        return self.__name

    @name.setter
    def name(self, value):
        """Sets __name to value.

        :param value: New name
        :type value: str

        """

        self.__name = value

    @property
    def identifier(self):
        """Returns the feature's local identifier.

        :returns: Identifier
        :rtype: str

        """

        return self.__identifier

    @identifier.setter
    def identifier(self, value):
        """Builds an ID object from value and assigns the resulting identifier
        to __identifier.

        :param value: Proposal of the identifier
        :type value: str
        """

        self.__identifier = value

    @property
    def subFeatures(self):
        """Returns a dict containing all subordinate features.

        :returns: All subordinate features.
        :rtype: dict[<str>, <Feature>]

        """

        return self.__subFeatures

    @subFeatures.setter
    def subFeatures(self, value):
        """Replaces __subFeatures with value.

        :param value: New collection of subordinate features

        """

        # !!! Do we really want this? Better add and remove functions.
        self.__subFeatures = value

    @property
    def json_out(self):
        return self.__json_out

    def to_json(self):
        """Returns a JSON representation of this feature. Note that this
        function works recursively.

        :returns: Json representation
        :rtype: str

        """

        self.__json_out["class"] = self.__class_name

        if self.identifier:
            self.__json_out["identifier"] = self.identifier
        if self.name:
            self.__json_out["name"] = self.name
        if self.subFeatures:
            self.__json_out["subFeatures"] = []
        for key in self.subFeatures.keys():
            res = self.subFeatures[key].to_json()
            self.__json_out["subFeatures"].append(res)
        return self.__json_out

    def get_my_thing(self):
        tmp = self.parent
        print("Searching for thing....")
        print(type(tmp))
        while not isinstance(tmp, Thing):
            if tmp is None:
                return None
            tmp = tmp.parent
        return tmp



from ml.ml40.features.functionalities.functionality import Functionality


class AcceptsWinchCommands(Functionality):
    """This functionality signalizes that the thing's winch can be moved remotely
    via S3I-B messages ."""

    def __init__(self, namespace="fml40", name="", identifier="", parent=None):
        """Initializes the object.

        :param name:  Object name
        :param identifier: Identifier

        """
        super().__init__(
            namespace=namespace, name=name, identifier=identifier, parent=parent
        )

    def rollDown(self, speed):
        """roll down the winch with specified speed

        :param speed: speed
        :type speed: float
        """
        print("roll down the winch with the speed {}".format(speed))

    def rollUp(self, speed):
        """
        roll up the winch with specified speed

        :param speed: speed
        :type speed: float
        """
        print("roll up the winch with the speed {}".format(speed))



"""This module provides a logger."""

import logging
import os

APP_LOGGER = logging.getLogger("app_logger")


def setup_logger(dt_name, app_logger=APP_LOGGER):
    """Creates logger named app_logger.

    Logs are printed to stdout and saved to log files under '/logs'.
    Each file is names 'dt_name'.log.
    :param dt_name: Name of the digital twin.
    :returns: Created logger
    :rtype: Logger

    """
    if not os.path.exists("logs"):
        os.mkdir("logs")

    app_logger.setLevel(logging.DEBUG)

    log_formatter = logging.Formatter(
        "%(asctime)s - %(levelname)s - {}: %(message)s".format(dt_name)
    )

    stream_handler = logging.StreamHandler()
    stream_handler.setFormatter(log_formatter)
    # stream_handler.setLevel(logging.DEBUG)
    file_handler = logging.FileHandler(filename="./logs/{}.log".format(dt_name))
    file_handler.setLevel(logging.DEBUG)
    file_handler.setFormatter(log_formatter)
    app_logger.addHandler(file_handler)
    app_logger.addHandler(stream_handler)
    return app_logger



import pprint
import re

from ml.dt_factory import DT_FACTORY, RE_NAMESPACE
from ml.feature import Feature
from ml.role import Role


def test_namespace_re():
    assert not RE_NAMESPACE.fullmatch("")
    assert not RE_NAMESPACE.fullmatch("a")
    assert not RE_NAMESPACE.fullmatch("2")
    assert not RE_NAMESPACE.fullmatch("a:b")
    assert not RE_NAMESPACE.fullmatch("1:b")
    assert not RE_NAMESPACE.fullmatch("b:1")
    assert RE_NAMESPACE.fullmatch("a::b")
    assert not RE_NAMESPACE.fullmatch("::b")
    assert not RE_NAMESPACE.fullmatch("a::")
    assert RE_NAMESPACE.fullmatch("ba::aa")
    assert not RE_NAMESPACE.fullmatch("b::a::a")
    assert not RE_NAMESPACE.fullmatch("b::a::1")


def test_class_instantiation():
    for key, s3i_class in DT_FACTORY.items():
        s3i_inst = s3i_class()
        assert RE_NAMESPACE.fullmatch(s3i_inst.class_name)



from ml.ml40.features.properties.values.documents.jobs.job import Job


class LogTransportationJob(Job):
    def __init__(self, namespace="fml40", name="", identifier=""):
        super().__init__(namespace=namespace, name=name, identifier=identifier)
        self.__woodPiles = []

    @property
    def woodPiles(self):
        return self.__woodPiles

    @woodPiles.setter
    def woodPiles(self, value):
        self.__woodPiles.append(value)



from ml.ml40.features.properties.values.value import Value


class HarvestedVolume(Value):
    def __init__(self, namespace="fml40", name="", identifier="", parent=None):
        super().__init__(namespace=namespace, name=name, identifier=identifier, parent=parent)

        self.__volume = None
        self.__json_out = None

    @property
    def volume(self):
        return self.__volume

    @volume.setter
    def volume(self, value):
        self.__volume = value

    def to_json(self):
        self.__json_out = super().to_json()
        if self.volume is not None:
            self.__json_out["volume"] = self.volume
        return self.__json_out



from ml.ml40.features.properties.values.value import Value


class DominantTreeSpecies(Value):
    def __init__(self, namespace="fml40", name="", identifier="", parent=None):
        super().__init__(
            namespace=namespace, name=name, identifier=identifier, parent=parent
        )



from ml.ml40.features.properties.values.value import Value


class FellingPeriod(Value):
    def __init__(self, namespace="fml40", name="", identifier="", parent=None):
        super().__init__(
            namespace=namespace, name=name, identifier=identifier, parent=parent
        )
        self.__from = None
        self.__to = None
        self.__json_out = dict()

    @property
    def _from(self):
        return self.__from

    @_from.setter
    def _from(self, value):
        self.__from = value

    @property
    def to(self):
        return self.__to

    @to.setter
    def to(self, value):
        self.__to = value

    def to_json(self):
        self.__json_out = super().to_json()
        if self._from is not None:
            self.__json_out["from"] = self.__from
        if self.to is not None:
            self.__json_out["to"] = self.__to
        return self.__json_out



from ml.ml40.features.functionalities.functionality import Functionality
from ml.ml40.features.properties.values.location import Location
from ml.ml40.features.properties.values.route import Route


class PlansRoutes(Functionality):
    def __init__(self, namespace="ml40", name="", identifier="", parent=None):
        super().__init__(
            namespace=namespace, name=name, identifier=identifier, parent=parent
        )

    def planRoute(self, start: Location, goal: Location) -> Route:
        pass



"""This module implements the class AcceptsFellingJobs."""

from ml.fml40.features.properties.values.documents.jobs.felling_job import FellingJob
from ml.ml40.features.functionalities.accepts_jobs import AcceptsJobs
from ml.ml40.features.properties.values.documents.jobs.job_status import JobStatus


class AcceptsFellingJobs(AcceptsJobs):
    """This functionality signalizes that FellingJobs can be processed and
    offers the possibility to remove and query it."""

    def __init__(self, namespace="fml40", name="", identifier="", parent=None):
        """Initializes the object.

        :param name: Object name
        :param identifier: Identifier

        """
        super().__init__(
            namespace=namespace, name=name, identifier=identifier, parent=parent
        )

    def acceptJob(self, job: FellingJob) -> bool:
        """Accepts the given FellingJob job. Returns true if the job has been
        accepted, otherwise returns false.

        :param job: FellingJob to be accepted
        :rtype: bool

        """
        pass

    def queryJobStatus(self, identifier) -> JobStatus:
        """Returns the status of the job whose id is identical to identifier.

        :param identifier: Identifier of the job to be queried
        :returns: Status of the job
        :rtype: JobStatus

        """
        pass

    def removeJob(self, identifier) -> bool:
        """Removes the job with the id identical to identifier. Returns true
        if the job has been removed, otherwise returns false.

        :param identifier: Identifier of the job to be deleted
        :rtype: bool

        """
        pass



from ml.ml40.features.properties.values.value import Value


class HeartRate(Value):
    def __init__(self, namespace="ml40", name="", identifier=""):
        super().__init__(
            namespace=namespace,
            name=name,
            identifier=identifier
        )
        self.__current_heart_rate = None
        self.__min_heart_rate = None
        self.__max_heart_rate = None
        self.__mean_heart_rate = None

    @property
    def currentHeartRate(self):
        return self.__current_heart_rate

    @currentHeartRate.setter
    def currentHeartRate(self, value):
        self.__current_heart_rate = value

    @property
    def minHeartRate(self):
        return self.__min_heart_rate

    @minHeartRate.setter
    def minHeartRate(self, value):
        self.__min_heart_rate = value

    @property
    def maxHeartRate(self):
        return self.__max_heart_rate

    @maxHeartRate.setter
    def maxHeartRate(self, value):
        self.__max_heart_rate = value

    @property
    def meanHeartRate(self):
        return self.__mean_heart_rate

    @meanHeartRate.setter
    def meanHeartRate(self, value):
        self.__mean_heart_rate = value

    def to_json(self):
        self.__json_out = super().to_json()
        if self.__current_heart_rate is not None:
            self.__json_out["currentHeartRate"] = self.__current_heart_rate

        if self.__min_heart_rate is not None:
            self.__json_out["minHeartRate"] = self.__min_heart_rate

        if self.__max_heart_rate is not None:
            self.__json_out["maxHeartRate"] = self.__max_heart_rate

        if self.__mean_heart_rate is not None:
            self.__json_out["meanHeartRate"] = self.__mean_heart_rate

        return self.__json_out



class Parameters(object):
    DEFAULT_IDP_URL = "https://idp.s3i.vswf.dev/"
    DEFAULT_IDP_REALM = "KWH"
    DEFAULT_BROKER_HOST = "rabbitmq.s3i.vswf.dev"
    DEFAULT_BROKER_VHOST = "s3i"
    DEFAULT_REPO_WWS_URL = "wss://ditto.s3i.vswf.dev/ws/2"
    DEFAULT_REPO_URL = "https://ditto.s3i.vswf.dev/api/2/"
    DEFAULT_DIR_URL = "https://dir.s3i.vswf.dev/api/2/"
    DEFAULT_REPO_SNYC_FREQ = 1  # Hz
    DEFAULT_DIR_SYNC_FREQ = 0
    DEFAULT_THING_SYNC_FREQ = 50
    BROKER_MSG_COMPRESS_THRESHOLD = 1000000 # Bytes

    def __init__(self,
                 idp_url=None,
                 idp_realm=None,
                 broker_host=None,
                 broker_vhost=None,
                 repo_url=None,
                 repo_wws_url=None,
                 dir_url=None,
                 repo_sync_freq=None,
                 dir_sync_freq=None,
                 thing_sync_freq=None,
                 broker_msg_compress_threshold=None
                 ):

        self.__idp_url = idp_url
        self.__idp_realm =idp_realm
        self.__broker_host = broker_host
        self.__broker_vhost = broker_vhost
        self.__repo_url = repo_url
        self.__repo_wws_url = repo_wws_url
        self.__dir_url = dir_url
        self.__repo_sync_freq = repo_sync_freq
        self.__dir_sync_freq = dir_sync_freq
        self.__thing_sync_freq = thing_sync_freq
        self.__broker_msg_compress_threshold = broker_msg_compress_threshold

        if self.__idp_url is None:
            self.__idp_url = self.DEFAULT_IDP_URL

        if self.__idp_realm is None:
            self.__idp_realm = self.DEFAULT_IDP_REALM

        if self.__broker_host is None:
            self.__broker_host = self.DEFAULT_BROKER_HOST

        if self.__broker_vhost is None:
            self.__broker_vhost = self.DEFAULT_BROKER_VHOST

        if self.__repo_url is None:
            self.__repo_url = self.DEFAULT_REPO_URL

        if self.__repo_wws_url is None:
            self.__repo_wws_url = self.DEFAULT_REPO_WWS_URL

        if self.__dir_url is None:
            self.__dir_url = self.DEFAULT_DIR_URL

        if self.__repo_url is None:
            self.__repo_url = self.DEFAULT_REPO_URL

        if self.__repo_sync_freq is None:
            self.__repo_sync_freq = self.DEFAULT_REPO_SNYC_FREQ

        if self.__dir_sync_freq is None:
            self.__dir_sync_freq = self.DEFAULT_DIR_SYNC_FREQ

        if self.__thing_sync_freq is None:
            self.__thing_sync_freq = self.DEFAULT_THING_SYNC_FREQ

        if self.__broker_msg_compress_threshold is None:
            self.__broker_msg_compress_threshold = self.BROKER_MSG_COMPRESS_THRESHOLD

    @property
    def idp_url(self):
        return self.__idp_url

    @idp_url.setter
    def idp_url(self, value):
        self.__idp_url = value

    @property
    def idp_realm(self):
        return self.__idp_realm

    @idp_realm.setter
    def idp_realm(self, value):
        self.__idp_realm = value

    @property
    def broker_host(self):
        return self.__broker_host

    @broker_host.setter
    def broker_host(self, value):
        self.__broker_host = value

    @property
    def broker_vhost(self):
        return self.__broker_vhost

    @broker_vhost.setter
    def broker_vhost(self, value):
        self.__broker_vhost = value

    @property
    def repo_url(self):
        return self.__repo_url

    @repo_url.setter
    def repo_url(self, value):
        self.__repo_url = value

    @property
    def repo_wws_url(self):
        return self.__repo_wws_url

    @repo_wws_url.setter
    def repo_wws_url(self, value):
        self.__repo_wws_url = value

    @property
    def dir_url(self):
        return self.__dir_url

    @dir_url.setter
    def dir_url(self, value):
        self.__dir_url = value

    @property
    def repo_sync_freq(self):
        return self.__repo_sync_freq

    @repo_sync_freq.setter
    def repo_sync_freq(self, value):
        self.__repo_sync_freq = value

    @property
    def dir_sync_freq(self):
        return self.__dir_sync_freq

    @dir_sync_freq.setter
    def dir_sync_freq(self, value):
        self.__dir_sync_freq = value

    @property
    def thing_sync_freq(self):
        return self.__thing_sync_freq

    @thing_sync_freq.setter
    def thing_sync_freq(self, value):
        self.__thing_sync_freq = value

    @property
    def broker_msg_compress_threshold(self):
        return self.__broker_msg_compress_threshold

    @broker_msg_compress_threshold.setter
    def broker_msg_compress_threshold(self, value):
        self.__broker_msg_compress_threshold = value


from ml.ml40.features.properties.values.value import Value


class StdDeviationRPY(Value):
    def __init__(self, namespace="ml40", name="", identifier=""):
        super().__init__(
            namespace=namespace,
            name=name,
            identifier=identifier
        )

        self.__std_deviation_roll = None
        self.__std_deviation_pitch = None
        self.__std_deviation_yaw = None
        self.__json_out = {}

    @property
    def stdDeviationRoll(self):
        return self.__std_deviation_roll

    @stdDeviationRoll.setter
    def stdDeviationRoll(self, value):
        self.__std_deviation_roll = value

    @property
    def stdDeviationPitch(self):
        return self.__std_deviation_pitch

    @stdDeviationPitch.setter
    def stdDeviationPitch(self, value):
        self.__std_deviation_pitch = value

    @property
    def stdDeviationYaw(self):
        return self.__std_deviation_yaw

    @stdDeviationYaw.setter
    def stdDeviationYaw(self, value):
        self.__std_deviation_yaw = value

    def to_json(self):
        self.__json_out = super().to_json()
        if self.__std_deviation_roll is not None:
            self.__json_out["stdDeviationRoll"] = self.__std_deviation_roll
        if self.__std_deviation_yaw is not None:
            self.__json_out["stdDeviationPitch"] = self.__std_deviation_yaw
        if self.__std_deviation_pitch is not None:
            self.__json_out["stdDeviationYaw"] = self.__std_deviation_pitch

        return self.__json_out




"""This module implements the class AcceptsForwardingJobs."""

from ml.app_logger import APP_LOGGER
from ml.fml40.features.properties.values.documents.jobs.forwarding_job import (
    ForwardingJob,
)
from ml.ml40.features.functionalities.accepts_jobs import AcceptsJobs


class AcceptsForwardingJobs(AcceptsJobs):
    """This functionality signalizes that ForwadingJobs can be processed."""

    def __init__(self, namespace="fml40", name="", identifier="", parent=None):
        """Initializes the object.

        :param name:  Object name
        :param identifier: Identifier

        """
        super().__init__(
            namespace=namespace, name=name, identifier=identifier, parent=parent
        )

    def acceptJob(self, job: ForwardingJob) -> bool:
        """Accepts the given ForwardingJob. Returns true if the job has been
        accepted, otherwise returns false.

        :param job: ForwardingJob to be accepted
        :rtype: bool

        """
        pass



from ml.feature import Feature


class Functionality(Feature):
    """Genric implementation of a Functionaliy."""

    def __init__(self, namespace="ml40", name="", identifier="", parent=None):
        super(Functionality, self).__init__(
            namespace=namespace, name=name, identifier=identifier, parent=parent
        )



from ml.ml40.roles.hmis.hmi import HMI


class App(HMI):
    def __init__(self, namespace="ml40", name="", identifier="", parent=None):
        super(App, self).__init__(
            namespace=namespace, name=name, identifier=identifier, parent=parent
        )



"""This module implements the class Fells."""

from ml.fml40.features.properties.values.documents.jobs.felling_job import \
    FellingJob
from ml.fml40.features.properties.values.tree_data import TreeData
from ml.ml40.features.functionalities.functionality import Functionality


class Fells(Functionality):
    """This functionality signalizes that FellingJobs can be processed and
    offers the possibility to fell trees."""

    def __init__(self, namespace="fml40", name="", identifier="", parent=None):
        """Initializes the object.

        :param name: Object name
        :param identifier: Identifier

        """
        super().__init__(
            namespace=namespace, name=name, identifier=identifier, parent=parent
        )

    def executeFellingJob(self, job: FellingJob):
        """Executes the given FellingJob.

        :param FellingJob: Specification of the job
        """
        print("i am executing the felling job {}".format(job))
        pass

    def fell(self, tree_data: TreeData):
        """Fells the tree specified by tree_data

        :param tree_data: Description of the tree.
        """
        print("i am felling the tree {}".format(tree_data))
        pass



from abc import ABC


class ditto_feature(ABC):
    def __init__(self, id, key, value):
        self.__id = id
        self.__key = key
        self.__value = value

    @property
    def id(self):
        return self.__id

    @id.setter
    def id(self, value):
        self.__id = value

    @property
    def key(self):
        return self.__key

    @key.setter
    def key(self, value):
        self.__key = value

    @property
    def value(self):
        return self.__value

    @value.setter
    def value(self, value):
        self.__value = value

    def to_json(self):
        return {
            "properties": {
                self.__key: self.__value
            }
        }


from ml.ml40.features.properties.values.value import Value


class Weight(Value):
    def __init__(self, namespace="ml40", name="", identifier="", parent=None):
        super().__init__(
            namespace=namespace, name=name, identifier=identifier, parent=parent
        )

        self.__weight = None
        self.__json_out = {}

    @property
    def weight(self):
        return self.__weight

    @weight.setter
    def weight(self, value):
        self.__weight = value

    def to_json(self):
        self.__json_out = super().to_json()
        if self.weight is not None:
            self.__json_out["weight"] = self.weight
        return self.__json_out



"""This module implements the class AcceptsLogMeasurements."""

from ml.fml40.features.properties.values.documents.reports.log_measurement import (
    LogMeasurement,
)
from ml.ml40.features.functionalities.functionality import Functionality


class AcceptsLogMeasurements(Functionality):
    """This functionality signalizes that LogMeasurements can be processed."""

    def __init__(self, namespace="fml40", name="", identifier="", parent=None):
        """Initializes the object.

        :param name:  Object name
        :param identifier: Identifier

        """
        super().__init__(
            namespace=namespace, name=name, identifier=identifier, parent=parent
        )

    def acceptLogMeasurement(self, log_measurement: LogMeasurement) -> bool:
        """Accepts the given LogMeasurement. Returns true if the job has been
        accepted, otherwise returns false.

        :param job: LogMeasurement to be accepted
        :rtype: bool

        """
        pass



from ml.ml40.features.properties.values.value import Value


class Stretch(Value):
    def __init__(self, namespace="mml40", name="", identifier=""):
        super().__init__(namespace=namespace, name=name, identifier=identifier)

        self.__stretch = None
        self.__min_stretch = None
        self.__max_stretch = None

    @property
    def stretch(self):
        return self.__stretch

    @stretch.setter
    def stretch(self, value):
        self.__stretch = value

    @property
    def minStretch(self):
        return self.__min_stretch

    @minStretch.setter
    def minStretch(self, value):
        self.__min_stretch = value

    @property
    def maxStretch(self):
        return self.__max_stretch

    @maxStretch.setter
    def maxStretch(self, value):
        self.__max_stretch = value

    def to_json(self):
        self.__json_out = super().to_json()
        if self.__stretch is not None:
            self.__json_out["stretch"] = self.__stretch
        if self.__min_stretch is not None:
            self.__json_out["minStretch"] = self.__min_stretch
        if self.__max_stretch is not None:
            self.__json_out["maxStretch"] = self.__max_stretch
        return self.__json_out



from ml.ml40.features.properties.values.value import Value


class OrientationRPY(Value):
    def __init__(self, namespace="ml40", name="", identifier="", parent=None):
        super().__init__(
            namespace=namespace,
            name=name,
            identifier=identifier,
            parent=parent
        )

        self.__roll = None
        self.__pitch = None
        self.__yaw = None
        self.__json_out = dict()

    @property
    def roll(self):
        return self.__roll

    @roll.setter
    def roll(self, value):
        self.__roll = value

    @property
    def pitch(self):
        return self.__pitch

    @pitch.setter
    def pitch(self, value):
        self.__pitch = value

    @property
    def yaw(self):
        return self.__yaw

    @yaw.setter
    def yaw(self, value):
        self.__yaw = value

    def to_json(self):
        self.__json_out = super().to_json()
        if self.__roll is not None:
            self.__json_out["roll"] = self.__roll
        if self.__yaw is not None:
            self.__json_out["yaw"] = self.__yaw
        if self.__pitch is not None:
            self.__json_out["pitch"] = self.__pitch

        return self.__json_out




from ml.ml40.features.properties.values.value import Value


class TimeSlot(Value):
    def __init__(self, namespace="ml40", name="", identifier="", parent=None):
        super().__init__(
            namespace=namespace, name=name, identifier=identifier, parent=parent
        )
        self.end = None
        self.start = None
        self.__json_out = dict()

    def to_json(self):
        self.__json_out = super().to_json()
        if self.end is not None:
            self.__json_out["end"] = self.end
        if self.start is not None:
            self.__json_out["start"] = self.start
        return self.__json_out



from ml.ml40.features.properties.values.value import Value


class Location(Value):
    def __init__(self, namespace="ml40", name="", identifier="", parent=None):
        super().__init__(
            namespace=namespace, name=name, identifier=identifier, parent=parent
        )
        self.__latitude = None
        self.__longitude = None
        self.__orientation = None
        self.__json_out = {}

    @property
    def latitude(self):
        return self.__latitude

    @latitude.setter
    def latitude(self, value):
        self.__latitude = value

    @property
    def longitude(self):
        return self.__longitude

    @longitude.setter
    def longitude(self, value):
        self.__longitude = value

    @property
    def orientation(self):
        return self.__orientation

    @orientation.setter
    def orientation(self, value):
        self.__orientation = value

    def to_json(self):
        self.__json_out = super().to_json()
        if self.latitude is not None:
            self.__json_out["latitude"] = self.latitude
        if self.longitude is not None:
            self.__json_out["longitude"] = self.longitude
        if self.orientation is not None:
            self.__json_out["orientation"] = self.orientation
        return self.__json_out



"""This module implements the class ClassifiesTreeSpecies."""

from ml.ml40.features.functionalities.functionality import Functionality


class ClassifiesTreeSpecies(Functionality):
    """This functionality allows the classification of tree species."""

    def __init__(self, namespace="fml40", name="", identifier="", parent=None):
        """Initializes the object.

        :param name:  Object name
        :param identifier: Identifier

        """
        super().__init__(
            namespace=namespace, name=name, identifier=identifier, parent=parent
        )

    def calculateTreeSpeciesClassification(self, tree: bytes) -> bytes:
        """Returns the classification of the species of tree.

        :param: tree: Specification of tree
        :return: Tree species
        :rtype: bytes
        """
        pass



from ml.fml40.features.properties.values.tree_data import TreeData
from ml.identifier import ID
from ml.ml40.features.functionalities.functionality import Functionality
from ml.ml40.features.properties.values.location import Location


class ProvidesTreeData(Functionality):
    def __init__(self, namespace="fml40", name="", identifier="", parent=None):
        super().__init__(
            namespace=namespace, name=name, identifier=identifier, parent=parent
        )

    def getTreeData(self, Tree: ID) -> list:
        pass

    def getTreesInDiameter(self, location: Location, diameter: float) -> list:
        pass



from ml.identifier import ID
from ml.ml40.features.functionalities.functionality import Functionality


class Renders(Functionality):
    def __init__(self, namespace="ml40", name="", identifier="", parent=None):
        super().__init__(
            namespace=namespace, name=name, identifier=identifier, parent=parent
        )

    def create3DVideo(self, identifier: ID) -> bytes:
        pass



from ml.ml40.features.properties.associations.association import Association


class Shared(Association):
    def __init__(self, namespace="ml40", name="", identifier="", parent=None):
        super().__init__(
            namespace=namespace, name=name, identifier=identifier, parent=parent
        )
        self.__targets = []
        self.__targets = dict()
        self.__json_out = dict()

    @property
    def targets(self):
        return self.__targets

    @targets.setter
    def targets(self, value):
        self.__targets = value

    def to_json(self):
        self.__json_out = super().to_json()
        if self.targets:
            self.__json_out["targets"] = []
            for key in self.__targets.keys():
                self.__json_out["targets"].append(self.targets[key].entry.refresh_sub_thing_repo_entry())
        return self.__json_out



from ml.ml40.features.properties.property import Property


class Value(Property):
    def __init__(self, namespace="ml40", name="", identifier="", parent=None):
        super().__init__(
            namespace=namespace, name=name, identifier=identifier, parent=parent
        )
        self.__valid_to = ""
        self.__valid_from = ""
        # TODO validate the datetype of valid from and valid to
        self.__json_out = {}

    @property
    def valid_to(self):
        return self.__valid_to

    @valid_to.setter
    def valid_to(self, value):
        self.__valid_to = value

    @property
    def valid_from(self):
        return self.__valid_from

    def valid_from(self, value):
        self.__valid_from = value

    def to_json(self):
        self.__json_out = super().to_json()
        if self.valid_from and self.valid_to:
            self.__json_out["valid_from"] = self.valid_from
            self.__json_out["valid_to"] = self.valid_to
        return self.__json_out



from ml.ml40.features.properties.values.value import Value


class Address(Value):
    def __init__(self, namespace="ml40", name="", identifier="", parent=None):
        super().__init__(
            namespace=namespace, name=name, identifier=identifier, parent=parent
        )

        self.city = None
        self.country = None
        self.street = None
        self.streetnumber = None
        self.zip = None
        self.__json_out = dict()

    def to_json(self):
        self.__json_out = super().to_json()
        if self.city is not None:
            self.__json_out["city"] = self.city
        if self.country is not None:
            self.__json_out["country"] = self.country
        if self.street is not None:
            self.__json_out["street"] = self.street
        if self.streetnumber is not None:
            self.__json_out["streetnumber"] = self.streetnumber
        if self.zip is not None:
            self.__json_out["zip"] = self.zip

        return self.__json_out



from ml.ml40.features.properties.values.value import Value


class Temperature(Value):
    def __init__(self, namespace="ml40", name="", identifier="", parent=None):
        super().__init__(
            namespace=namespace, name=name, identifier=identifier, parent=parent
        )
        self.__temperature = None
        self.__json_out = dict()

    @property
    def temperature(self):
        return self.__temperature

    @temperature.setter
    def temperature(self, value):
        if isinstance(value, (float, int)):
            self.__temperature = value
        else:
            raise TypeError

    def to_json(self):
        self.__json_out = super().to_json()
        if self.temperature is not None:
            self.__json_out["temperature"] = self.temperature
        return self.__json_out



from ml.ml40.features.functionalities.functionality import Functionality
from ml.ml40.features.properties.values.documents.jobs.job import Job


class ClearsJobs(Functionality):
    def __init__(self, namespace="ml40", name="", identifier="", parent=None):
        super().__init__(
            namespace=namespace, name=name, identifier=identifier, parent=parent
        )

    def clear(self, job: Job) -> bool:
        pass



from ml.role import Role


class DT(Role):
    def __init__(self, namespace="ml40", name="", identifier="", parent=None):
        super().__init__(namespace=namespace, name=name, identifier=identifier, parent=parent)



from ml.ml40.roles.dts.sensors.sensor import Sensor


class StrainGauge(Sensor):
    def __init__(self, namespace="mml40", name="", identifier=""):
        super().__init__(namespace=namespace, name=name, identifier=identifier)



"""This module implements the class EvaluatesStandAttributes."""

from datetime import date

from ml.ml40.features.functionalities.functionality import Functionality


class EvaluatesStandAttributes(Functionality):
    def __init__(self, namespace="fml40", name="", identifier="", parent=None):
        """Initializes the object.

        :param name: Object name
        :param identifier: Identifier

        """
        super().__init__(
            namespace=namespace, name=name, identifier=identifier, parent=parent
        )

    def calculateStandAttributes(self, input_a: bytes, input_b: date) -> str:
        pass

    def calculateStock(self, input_a: bytes) -> float:
        pass



import asyncio


class CallbackManager(object):
    """
    Global callback system, which is aimed at to be a single place to manage callbacks and process them
    """
    CALLBACK = 'callback'
    ONE_SHOT = 'one_shot'
    IS_ASYNC_CALLBACK = 'is_async_callback'
    ARGS = 'args'
    KWARGS = 'kwargs'

    def __init__(self):
        """
        Create an instance of the CallbackManager
        """
        self._stack = {}

    def add(self, prefix, callback, one_shot, is_async, *args, **kwargs):
        """
        Appends async callback function to dictionary
        :param prefix:
        :param callback:
        :param one_shot:
        :param is_async:
        :param args:
        :param kwargs:
        :return:
        """
        # prepare the stack
        if prefix not in self._stack:
            self._stack[prefix] = []

        if not isinstance(one_shot, bool):
            raise TypeError
        if not isinstance(is_async, bool):
            raise TypeError

        # create callback dictionary
        callback_dict = self.create_callback_dict(
            callback,
            one_shot,
            is_async,
            *args,
            **kwargs
        )

        # check for a duplicate
        if callback_dict in self._stack[prefix]:
            return prefix

        # append
        self._stack[prefix].append(callback_dict)
        return prefix

    def remove(self, prefix, callback_dict):
        if prefix not in self._stack:
            return False
        for callback in self._stack[prefix]:
            if callback == callback_dict:
                self._stack[prefix].remove(callback)
                return True
        return False

    def clear(self):
        if self._stack:
            self._stack = {}

    def process(self, prefix, loop=asyncio.get_event_loop()):
        if prefix not in self._stack:
            return False

        for callback_dict in self._stack[prefix]:
            method = callback_dict[self.CALLBACK]
            args = callback_dict[self.ARGS]
            kwargs = callback_dict[self.KWARGS]
            if callback_dict[self.IS_ASYNC_CALLBACK]:
                loop.create_task(method(*args, **kwargs))
            else:
                method(*args, **kwargs)

        for callback_dict in self._stack[prefix]:
            if callback_dict[self.ONE_SHOT]:
                self.remove(prefix, callback_dict)

        return True

    def create_callback_dict(self, callback, one_shot, is_async, *args, **kwargs):
        """
        Create and return callback dictionary

        :param method callback:
        :param bool one_shot:
        :param is_async:
        :return:
        """

        return {
            self.CALLBACK: callback,
            self.ONE_SHOT: one_shot,
            self.IS_ASYNC_CALLBACK: is_async,
            self.ARGS: args,
            self.KWARGS: kwargs
        }






"""This module implements the class AcceptsFellingSupportJobs."""

from ml.fml40.features.properties.values.documents.jobs.fellung_support_job import (
    FellingSupportJob,
)
from ml.ml40.features.functionalities.accepts_jobs import AcceptsJobs


class AcceptsFellingSupportJobs(AcceptsJobs):
    """This functionality signalizes that FellingSupportJobs can be
    processed."""

    def __init__(self, namespace="fml40", name="", identifier="", parent=None):
        """Initializes the object.

        :param name:  Object name
        :param identifier: Identifier

        """
        super().__init__(
            namespace=namespace, name=name, identifier=identifier, parent=parent
        )

    def acceptJob(self, job: FellingSupportJob) -> bool:
        """Accepts the given FellingSupportJob job. Returns true if the job
        has been accepted, otherwise returns false.

        :param job: FellingSupportJob to be accepted
        :rtype: bool
        """
        pass



from ml.identifier import ID
from ml.ml40.features.functionalities.functionality import Functionality
from ml.ml40.features.properties.values.documents.jobs.job import Job
from ml.ml40.features.properties.values.documents.reports.report import Report


class ManagesJobs(Functionality):
    def __init__(self, namespace="ml40", name="", identifier="", parent=None):
        super().__init__(
            namespace=namespace,
            name=name,
            identifier=identifier,
            parent=parent)

    def acceptReport(self, report: Report):
        pass

    def assignJob(self, job: Job, identifier: ID):
        pass



"""This module implements the class AcceptsProximityAlert."""

from ml.ml40.features.functionalities.functionality import Functionality


class AcceptsProximityAlert(Functionality):
    """This functionality signalizes that an alert is generated if things
    are to close to this thing."""

    def __init__(self, namespace="fml40", name="", identifier="", parent=None):
        """Initializes the object.

        :param name:  Object name
        :param identifier: Identifier

        """
        super().__init__(
            namespace=namespace, name=name, identifier=identifier, parent=parent
        )

    def proximityAlert(self, ids: list, distances: list):
        print("Making Proximity Alert...")



from ml.ml40.features.properties.values.value import Value


class GenericProperty(Value):
    def __init__(self, namespace="ml40", name="", identifier="", parent=None):
        super().__init__(
            namespace=namespace, name=name, identifier=identifier, parent=parent
        )
        self.value = None
        self.__json_out = dict()

    def to_json(self):
        self.__json_out = super().to_json()
        if self.value is not None:
            self.__json_out["value"] = self.value

        return self.__json_out



from ml.fml40.roles.dts.forest.forest import Forest


class ForestSegment(Forest):
    def __init__(self, namespace="fml40", name="", identifier="", parent=None):
        super().__init__(namespace=namespace, name=name, identifier=identifier, parent=parent)



from ml.ml40.features.properties.values.value import Value


class StemSegmentProperties(Value):
    def __init__(self, namespace="fml40", name="", identifier="", parent=None):
        super().__init__(namespace=namespace, name=name, identifier=identifier, parent=parent)

        self.__diameter = None
        self.__length = None
        self.__quality = None
        self.__woodType = None
        self.__json_out = dict()

    @property
    def diameter(self):
        return self.__diameter

    @diameter.setter
    def diameter(self, value):
        self.__diameter = value

    @property
    def length(self):
        return self.__length

    @length.setter
    def length(self, value):
        self.__length = value

    @property
    def quality(self):
        return self.__quality

    @quality.setter
    def quality(self, value):
        self.__quality = value

    @property
    def woodType(self):
        return self.__woodType

    @woodType.setter
    def woodType(self, value):
        self.__woodType = value

    def to_json(self):
        self.__json_out = super().to_json()
        if self.diameter is not None:
            self.__json_out["diameter"] = self.diameter

        if self.length is not None:
            self.__json_out["length"] = self.length

        if self.woodType is not None:
            self.__json_out["woodType"] = self.woodType

        if self.quality is not None:
            self.__json_out["quality"] = self.quality

        return self.__json_out



from ml.ml40.roles.dts.sites.site import Site


class Mill(Site):
    def __init__(self, namespace="fml40", name="", identifier=""):
        super().__init__(namespace=namespace, name=name, identifier=identifier)



from ml.ml40.roles.dts.sensors.sensor import Sensor


class IMU(Sensor):
    """
    Inertia measurement unit
    """
    def __init__(self, namespace="ml40", name="", identifier=""):
        super(IMU, self).__init__(
            namespace=namespace, name=name, identifier=identifier
        )



from ml.ml40.roles.dts.dt import DT


class Sensor(DT):
    def __init__(self, namespace="ml40", name="", identifier="", parent=None):
        super().__init__(
            namespace=namespace, name=name, identifier=identifier, parent=parent
        )



from ml.app_logger import APP_LOGGER
from ml.fml40.features.properties.values.documents.jobs.felling_job import FellingJob
from ml.ml40.features.functionalities.functionality import Functionality


class Harvests(Functionality):
    def __init__(self, namespace="fml40", name="", identifier="", parent=None):
        super().__init__(
            namespace=namespace,
            name=name,
            identifier=identifier,
        )

    def executeJob(self, job: FellingJob):
        pass



from ml.ml40.features.properties.values.value import Value


class Count(Value):
    def __init__(self, namespace="ml40", name="", identifier="", parent=None):
        super().__init__(
            namespace=namespace, name=name, identifier=identifier, parent=parent
        )
        self.currentCount = None
        self.maxCount = None
        self.minCount = None
        self.__json_out = dict()

    def to_json(self):
        self.__json_out = super().to_json()
        if self.currentCount is not None:
            self.__json_out["currentCount"] = self.currentCount
        if self.maxCount is not None:
            self.__json_out["maxCount"] = self.maxCount
        if self.minCount is not None:
            self.__json_out["minCount"] = self.minCount

        return self.__json_out



from ml.ml40.features.properties.values.value import Value


class FellIndicator(Value):
    def __init__(self, namespace="fml40", name="", identifier="", parent=None):
        super().__init__(namespace=namespace, name=name, identifier=identifier, parent=parent)
        self.__marked = None
        self.__json_out = dict()

    @property
    def marked(self):
        return self.__marked

    @marked.setter
    def marked(self, value):
        self.__marked = value

    def to_json(self):
        self.__json_out = super().to_json()
        if self.marked is not None:
            self.__json_out["marked"] = self.__marked
        return self.__json_out



from ml.ml40.features.properties.values.value import Value


class Distance(Value):
    def __init__(self, namespace="ml40", name="", identifier="", parent=None):
        super().__init__(
            namespace=namespace, name=name, identifier=identifier, parent=parent
        )
        self.__distance = None

    @property
    def distance(self):
        return self.__distance

    @distance.setter
    def distance(self, value):
        self.__distance = value

    def to_json(self):
        self.__json_out = super().to_json()
        if self.__distance is not None:
            self.__json_out["distance"] = self.__distance
        return self.__json_out



class Entry:
    def __init__(self, identifier, name):
        self.__name = name
        self.__identifier = identifier
        self.__features = {}
        self.__roles = {}
        self.__ditto_features = {}

        self.__dt_json = {}
        self.__repo_json = {}
        self.__dir_json = {
            "thingId": "",
            "policyId": "",
            "attributes": {

            }
        }

    @property
    def features(self):
        return self.__features

    @features.setter
    def features(self, value):
        self.__features = value

    @property
    def ditto_features(self):
        return self.__ditto_features

    @ditto_features.setter
    def ditto_features(self, value):
        self.__ditto_features = value

    @property
    def roles(self):
        return self.__roles

    @roles.setter
    def roles(self, value):
        self.__roles = value

    @property
    def name(self):
        return self.__name

    @property
    def identifier(self):
        return self.__identifier

    @property
    def dt_json(self):
        return self.__dt_json

    @property
    def repo_json(self):
        return self.__repo_json

    @property
    def dir_json(self):
        return self.__dir_json

    def refresh_directory_entry(self, current_dir_json):
        self.__dir_json = current_dir_json
        if self.__identifier is not None:
            self.__dir_json["thingId"] = self.__identifier
            self.__dir_json["policyId"] = self.__identifier
        if self.__name is not None:
            self.__dir_json["attributes"]["name"] = self.__name
        self.__dir_json["attributes"]["dataModel"] = "fml40"
        self.__dir_json["attributes"]["thingStructure"] = {
            "class": "ml40::Thing",
            "links": []
        }
        for key in self.__roles.keys():
            role_entry = {
                "association": "roles",
                "target": self.__roles[key].to_json()
            }
            self.__dir_json["attributes"]["thingStructure"]["links"].append(role_entry)

        for key in self.features.keys():
            feature_target = {
                "class": self.__features[key].to_json()["class"],
            }
            if self.__features[key].to_json().get("identifier") is not None:
                feature_target["identifier"] = self.__features[key].to_json()["identifier"]

            feature_entry = {"association": "features", "target": feature_target}
            # if the feature has targets, like ml40::Composite
            if hasattr(self.__features[key], "targets"):
                feature_entry["target"]["links"] = list()
                for target in self.__features[key].targets.keys():
                    target_json = (
                        self.__features[key].targets[target].entry.refresh_sub_thing_dir_entry()
                    )
                    feature_entry["target"]["links"].append(target_json)
            self.__dir_json["attributes"]["thingStructure"]["links"].append(feature_entry)
        return self.__dir_json

    def refresh_repository_entry(self):
        self.__repo_json = {
            "thingId": self.__identifier,
            "policyId": self.__identifier,
            "attributes": {
                "class": "ml40::Thing",
                "name": self.__name,
            },
        }
        if self.roles:
            self.__repo_json["attributes"]["roles"] = []
        if self.features:
            self.__repo_json["attributes"]["features"] = []
        if self.ditto_features:
            self.__repo_json["features"] = {}
        for key in self.__roles.keys():
            self.__repo_json["attributes"]["roles"].append(self.__roles[key].to_json())
        for key in self.__features.keys():
            self.__repo_json["attributes"]["features"].append(self.__features[key].to_json())
        for key in self.__ditto_features.keys():
            self.__repo_json["features"][key] = self.__ditto_features[key].to_json()
        return self.__repo_json

    def refresh_dt_json(self):
        self.__dt_json = self.refresh_repository_entry()

    def refresh_sub_thing_repo_entry(self):
        """Returns a dictionary representing this thing in it's current state
        as a subordinate thing. This representation should be used for
        subordinate things in s3i repository entries.

        :returns: Representation of this object as a subordinate thing
        :rtype: dict

        """

        json_out = {
            "class": "ml40::Thing",
            "name": self.name,
            "roles": [],
            "features": [],
        }
        if self.__identifier:
            json_out["identifier"] = self.__identifier
        for key in self.__roles.keys():
            json_out["roles"].append(self.__roles[key].to_json())
        for key in self.__features.keys():
            json_out["features"].append(self.__features[key].to_json())
        return json_out

    def refresh_sub_thing_dir_entry(self):
        """Returns a dictionary representing this thing in it's current state
        as a subordinate thing. This representation should be used for
        subordinate things in s3i directory entries.

        :returns: Representation of this object as a subordinate thing.
        :rtype: dict

        """

        json_out = {"class": "ml40::Thing", "links": []}
        if self.__identifier:
            json_out["identifier"] = self.__identifier
        for key in self.__roles.keys():
            role_entry = {"association": "roles", "target": self.__roles[key].to_json()}
            json_out["links"].append(role_entry)
        for key in self.__features.keys():
            feature_target = {
                "class": self.__features[key].to_json()["class"],
            }
            if self.__features[key].to_json().get("identifier") is not None:
                feature_target["identifier"] = self.__features[key].to_json()["identifier"]
            feature_entry = {"association": "features", "target": feature_target}
            json_out["links"].append(feature_entry)
        return json_out



from ml.ml40.features.properties.values.documents.jobs.job import Job


class ForwardingJob(Job):
    def __init__(self, namespace="fml40", name="", identifier=""):
        super().__init__(namespace=namespace, name=name, identifier=identifier)



from ml.ml40.features.functionalities.functionality import Functionality
from ml.ml40.features.properties.values.documents.reports.production_data import (
    ProductionData,
)


class ProvidesProductionData(Functionality):
    def __init__(self, namespace="fml40", name="", identifier="", parent=None):
        super().__init__(
            namespace=namespace, name=name, identifier=identifier, parent=parent
        )

    def getProductionData(self) -> ProductionData:
        return "Production Data is huge"



"""This module implements the class AcceptsMoveCommands."""

from ml.ml40.features.functionalities.functionality import Functionality


class AcceptsMoveCommands(Functionality):
    """This functionality signalizes that the thing can be moved remotely
    via S3I-B messages ."""

    def __init__(self, namespace="fml40", name="", identifier="", parent=None):
        """Initializes the object.

        :param name:  Object name
        :param identifier: Identifier

        """
        super().__init__(
            namespace=namespace, name=name, identifier=identifier, parent=parent
        )

    def move(self, longitude: float, latitude: float):
        """Moves the thing to the position specified longitude and latitude.

        :param longitude: Longitude
        :param latitude: Latitude

        """
        print("move to longitude: {}, latitude : {}".format(longitude, latitude))
        return {"longitude": longitude, "latitude": latitude}



from ml.ml40.features.properties.values.documents.document import Document
from ml.ml40.features.properties.values.documents.jobs.job_status import JobStatus


class Job(Document):
    def __init__(self, namespace="ml40", name="", identifier="", parent=None):
        super().__init__(
            namespace=namespace, name=name, identifier=identifier, parent=parent
        )
        self.__status = None
        self.__json_out = dict()

    @property
    def status(self):
        return self.__status

    @status.setter
    def status(self, value: JobStatus):
        self.__status = value

    def to_json(self):
        self.__json_out = super().to_json()
        if self.status is not None:
            self.__json_out["status"] = self.status
        return self.__json_out



"""This module implements the class Forwards."""

from ml.fml40.features.properties.values.documents.jobs.forwarding_job import (
    ForwardingJob,
)
from ml.ml40.features.functionalities.functionality import Functionality


class Forwards(Functionality):
    """This functionality signalizes that ForwardingJobs can be processed."""

    def __init__(self, namespace="fml40", name="", identifier="", parent=None):
        """Initializes the object.

        :param name: Object name
        :param identifier: Identifier

        """
        super().__init__(
            namespace=namespace, name=name, identifier=identifier, parent=parent
        )

    def executeJob(self, job):
        """Executes the given job.

        :param job: Job to be executed
        """
        pass

    def from_json(self, json_obj):
        """Initalizes this object from json_object.

        :param json_obj: JSON representation of this object
        """
        super().from_json(json_obj)



from ml.ml40.features.properties.values.value import Value


class Moisture(Value):
    def __init__(self, namespace="ml40", name="", identifier="", parent=None):
        super().__init__(
            namespace=namespace, name=name, identifier=identifier, parent=parent
        )
        self.__humidity = None
        self.__join_out = dict()

    @property
    def humidity(self):
        return self.__humidity

    @humidity.setter
    def humidity(self, value):
        self.__humidity = value

    def to_json(self):
        self.__join_out = super().to_json()
        if self.humidity is not None:
            self.__join_out["humidity"] = self.humidity
        return self.__join_out



from ml.ml40.features.properties.values.value import Value


class MaterialProperties(Value):
    def __init__(self, namespace="mml40", name="", identifier=""):
        super().__init__(namespace=namespace, name=name, identifier=identifier)
        self.__max_amount_load_change = None
        self.__material_type = None
        self.__yield_strength = None
        self.__young_modulus = None
        self.__json_out = dict()

    @property
    def maxAmountLoadChange(self):
        return self.__max_amount_load_change

    @maxAmountLoadChange.setter
    def maxAmountLoadChange(self, value):
        self.__max_amount_load_change = value

    @property
    def materialType(self):
        return self.__material_type

    @materialType.setter
    def materialType(self, value):
        self.__material_type = value

    @property
    def yieldStrength(self):
        return self.__yield_strength

    @yieldStrength.setter
    def yieldStrength(self, value):
        self.__yield_strength = value

    @property
    def youngModulus(self):
        return self.__young_modulus

    @youngModulus.setter
    def youngModulus(self, value):
        self.__young_modulus = value

    def to_json(self):
        self.__json_out = super().to_json()
        if self.maxAmountLoadChange is not None:
            self.__json_out["maxAmountLoadChange"] = self.maxAmountLoadChange

        if self.materialType is not None:
            self.__json_out["materialType"] = self.materialType

        if self.yieldStrength is not None:
            self.__json_out["yieldStrength"] = self.yieldStrength

        if self.youngModulus is not None:
            self.__json_out["youngModulus"] = self.youngModulus

        return self.__json_out



"""This module implements the class AcceptsMoistureMeasurement."""

from ml.fml40.features.properties.values.documents.reports.soil_moisture_measurement import (
    SoilMoistureMeasurement,
)
from ml.ml40.features.functionalities.functionality import Functionality


class AcceptsMoistureMeasurement(Functionality):
    """This functionality signalizes that SoilMoistureMeasurements can be
    processed."""

    def __init__(self, namespace="fml40", name="", identifier="", parent=None):
        """Initializes the object.

        :param name:  Object name
        :param identifier: Identifier

        """
        super().__init__(
            namespace=namespace, name=name, identifier=identifier, parent=parent
        )

    def acceptMoistureMeasurement(self, input: SoilMoistureMeasurement):
        """Accepts the given SoilMoistureMeasurement."""
        pass



from ml.ml40.features.functionalities.functionality import Functionality


class PredictsForestDevelopment(Functionality):
    def __init__(self, namespace="fml40", name="", identifier="", parent=None):
        super().__init__(namespace=namespace, name=name, identifier=identifier, parent=parent)



from ml.ml40.features.properties.values.value import Value


class LidarFrame(Value):
    def __init__(self, namespace="ml40", name="", identifier=""):
        super().__init__(
            namespace=namespace,
            name=name,
            identifier=identifier)

        self.__no_scan_lines = None
        self.__no_points = None

    @property
    def noScanLines(self):
        return self.__no_scan_lines

    @noScanLines.setter
    def noScanLines(self, value):
        self.__no_scan_lines = value

    @property
    def noPoints(self):
        return self.__no_points

    @noPoints.setter
    def noPoints(self, value):
        self.__no_points = value

    def to_json(self):
        self.__json_out = super().to_json()
        if self.__no_scan_lines is not None:
            self.__json_out["noScanLines"] = self.__no_scan_lines
        if self.noPoints is not None:
            self.__json_out["noPoints"] = self.noPoints
        return self.__json_out



import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="fml40-reference-implementation",
    version="0.2.4.1",
    author="Kompetenzzentrum Wald und Holz 4.0",
    author_email="s3i@kwh40.de",
    description="fml40 reference implementation basic functions",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://www.kwh40.de/",
    packages=setuptools.find_packages(),
    install_requires=[
        "requests",
        "jsonschema",
        "s3i==0.6.2",
        #"s3i@https://git.rwth-aachen.de/kwh40/s3i/-/jobs/artifacts/master/raw/public/s3i-0.5.3-py3-none-any.whl?job=wheel"
    ],
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: GNU Library or Lesser General Public License (LGPL)",
        "Operating System :: OS Independent",
    ]
)



"""This module implements the class AcceptsSingleTreeFellingJobs."""

from ml.ml40.features.functionalities.accepts_jobs import AcceptsJobs


class AcceptsSingleTreeFellingJobs(AcceptsJobs):
    """This functionality signalizes that single trees can be felled."""

    def __init__(self, namespace="fml40", name="", identifier="", parent=None):
        """Initializes the object.

        :param name:  Object name
        :param identifier: Identifier

        """
        super().__init__(
            namespace=namespace, name=name, identifier=identifier, parent=parent
        )



"""This module implements the class DeterminesPassability."""

from datetime import date

from ml.ml40.features.functionalities.functionality import Functionality


class DeterminesPassability(Functionality):
    """This functionality signalizes that the passability can be
    calculated."""

    def __init__(self, namespace="fml40", name="", identifier="", parent=None):
        """Initializes the object.

        :param name: Object name
        :param identifier: Identifier

        """
        super().__init__(
            namespace=namespace, name=name, identifier=identifier, parent=parent
        )

    def isPassableNow(self, input: float) -> bool:
        """Returns true if this object can be passed with the given weight
        input, otherwise returns false.

        :param input: Weight of the thing that should pass this object
        :rtype: bool
        """
        pass

    def maxPassableWeightNow(self) -> float:
        """Returns the maximal weight of an object that can pass this object
        at the moment.

        :returns: Maximal weight
        :rtype: bool
        """
        pass

    def maxPassableWeightOnDate(self, date: date) -> float:
        """Returns the maximal weight of an object that can pass this object
        at the given date.

        :param date: Specifies the date
        :returns: Maximal weight
        :rtype: bool
        """
        pass

    def willBePassable(self, input: float) -> date:
        """Returns the next date this object is passable.

        :param input: Weight that should pass this object
        :returns: Date
        :rtype: date
        """
        pass



from ml.ml40.features.properties.values.value import Value


class Tilt(Value):
    def __init__(self, namespace="fml40", name="", identifier="", parent=None):
        super().__init__(
            namespace=namespace, name=name, identifier=identifier, parent=parent
        )
        self.__direction = None
        self.__tilt = None
        self.__json_out = dict()

    @property
    def direction(self):
        return self.__direction

    @direction.setter
    def direction(self, value):
        self.__direction = value

    @property
    def tilt(self):
        return self.__tilt

    @tilt.setter
    def tilt(self, value):
        self.__tilt = value

    def to_json(self):
        self.__json_out = super().to_json()
        if self.direction is not None:
            self.__json_out["direction"] = self.direction

        if self.tilt is not None:
            self.__json_out["tilt"] = self.tilt

        return self.__json_out



from ml.ml40.features.properties.values.documents.reports.report import Report


class LogTransportationReport(Report):
    def __init__(self, namespace="fml40", name="", identifier=""):
        super().__init__(namespace=namespace, name=name, identifier=identifier)



from ml.ml40.roles.dts.handheld_devices.handheld_device import HandheldDevice


class Brushcutter(HandheldDevice):
    def __init__(self, namespace="fml40", name="", identifier="", parent=None):
        super().__init__(namespace=namespace, name=name, identifier=identifier, parent=parent)



from ml.ml40.roles.dts.sensors.sensor import Sensor


class LightSensor(Sensor):
    def __init__(self, namespace="ml40", name="", identifier="", parent=None):
        super(LightSensor, self).__init__(
            namespace=namespace, name=name, identifier=identifier, parent=parent
        )



from ml.ml40.features.properties.values.value import Value


class OperatingHours(Value):
    def __init__(self, namespace="ml40", name="", identifier="", parent=None):
        super().__init__(
            namespace=namespace, name=name, identifier=identifier, parent=parent
        )

        self.__total = None
        self.__json_out = dict()

    @property
    def total(self):
        return self.__total

    @total.setter
    def total(self, value):
        self.__total = value

    def to_json(self):
        self.__json_out = super().to_json()
        if self.__total is not None:
            self.__json_out["total"] = self.__total
        return self.__json_out



from ml.fml40.roles.dts.sites.mill.mill import Mill


class Papermill(Mill):
    def __init__(self, namespace="fml40", name="", identifier=""):
        super().__init__(namespace=namespace, name=name, identifier=identifier)



import asyncio
import json
import functools
import pika
import requests
from s3i.exception import raise_error_from_response, raise_error_from_s3ib_amqp, S3IBrokerRESTError, S3IBrokerAMQPError
from pika.adapters.asyncio_connection import AsyncioConnection
from pika.connection import ConnectionParameters
from pika import PlainCredentials
from pika.exceptions import UnroutableError
from ml.app_logger import APP_LOGGER
from ml.callback import CallbackManager
import ssl

MSG_EXCHANGE = "demo.direct"
EVENT_EXCHANGE = "eventExchange"
BROKER_HOST = "rabbitmq.s3i.vswf.dev"
BROKER_VIRTUAL_HOST = "s3i"
BROKER_API_URL = "https://broker.s3i.vswf.dev/"


class BrokerREST:
    def __init__(self, token, url=BROKER_API_URL):
        self.__token = token
        self.__url = url
        self.__headers = {
            'Content-Type': 'application/json',
            "Authorization": 'Bearer {}'.format(self.__token)
        }

    @property
    def token(self):
        return self.__token

    @token.setter
    def token(self, value):
        self.__token = value

    def send(self, endpoint, msg):
        response = requests.post(url=self.__url + endpoint, headers=self.__headers, data=msg)
        raise_error_from_response(response, S3IBrokerRESTError, 201)

    def receive_once(self, endpoint):
        response = requests.get(url=self.__url + endpoint,
                                headers=self.__headers)
        return raise_error_from_response(response, S3IBrokerRESTError, 200)


class Broker:
    _ON_CONNECTION_OPEN = "_on_connection_open"
    _ON_CONNECTION_CLOSED = "_on_connection_closed"
    _ON_CHANNEL_OPEN = "_on_channel_open"
    _ON_CHANNEL_CLOSED = "_on_channel_closed"

    def __init__(self, token, endpoint, callback, loop):
        self.__token = token
        self.__endpoint = endpoint #TODO Event Queue, more queue?
        self.__loop = loop
        self.__callback = callback
        self.__credentials = None
        self.__connection_parameters = None
        self.__connection = None
        self.__channel = None
        self.__consumer_tag = None

        self.__is_consuming = False
        self.__callbacks = CallbackManager()

    @property
    def token(self):
        return self.__token

    @token.setter
    def token(self, value):
        self.__token = value

    @property
    def connection(self):
        return self.__connection

    @property
    def channel(self):
        return self.__channel

    def connect(self):
        self.__credentials = PlainCredentials(
            username=" ",
            password=self.__token,
            erase_on_connect=True
        )
        self.__connection_parameters = ConnectionParameters(
            host=BROKER_HOST,
            virtual_host=BROKER_VIRTUAL_HOST,
            credentials=self.__credentials,
            heartbeat=10,
            port=5671,
            ssl_options=pika.SSLOptions(ssl.SSLContext())
        )

        self.__connection = AsyncioConnection(
            parameters=self.__connection_parameters,
            on_open_callback=self.on_connection_open,
            on_open_error_callback=self.on_connection_open_error,
            on_close_callback=self.on_connection_closed,
            custom_ioloop=self.__loop
        )

    def on_connection_open(self, _unused_connection):
        APP_LOGGER.info("[S3I]: Connection to Broker built")
        self.__channel = _unused_connection.channel(
            on_open_callback=self.on_channel_open
        )
        self.__callbacks.process(
            self._ON_CONNECTION_OPEN,
            self.__loop
        )

    @staticmethod
    def on_connection_open_error(_unused_connection, err):
        APP_LOGGER.error("[S3I]: Connection to broker failed: {}".format(err))

    def on_connection_closed(self, _unused_connection, reason):
        APP_LOGGER.info("[S3I]: Connection to Broker closed: {}".format(reason))
        if self.__is_consuming:
            self.__until_all_closed_and_reconnect()

    def on_channel_open(self, _unused_channel):
        APP_LOGGER.info("[S3I]: Channel open and start consuming messages")
        _unused_channel.add_on_close_callback(self.on_channel_closed)
        _unused_channel.basic_qos(
            prefetch_count=1
        )
        self.start_consuming()
        self.__callbacks.process(
            self._ON_CHANNEL_OPEN,
            self.__loop
        )

    def add_on_channel_open_callback(self, callback, one_shot, *args, **kwargs):
        self.__callbacks.add(
            self._ON_CHANNEL_OPEN,
            callback,
            one_shot,
            False,
            *args,
            **kwargs
        )

    def add_on_connection_open_callback(self, callback, one_shot, *args, **kwargs):
        self.__callbacks.add(
            self._ON_CONNECTION_OPEN,
            callback,
            one_shot,
            False,
            *args,
            **kwargs
        )

    def start_consuming(self):
        self.__consumer_tag = self.__channel.basic_consume(
            auto_ack=True,
            exclusive=True,
            queue=self.__endpoint,
            on_message_callback=self.__callback
        )
        self.__is_consuming = True

    def stop_consuming(self):
        cb = functools.partial(
            self.on_consumer_cancel_ok, userdata=self.__consumer_tag
        )
        self.__channel.basic_cancel(self.__consumer_tag, cb)
        self.__is_consuming = False

    def on_channel_closed(self, channel, reason):
        APP_LOGGER.info("[S3I]: Channel is closed: {}".format(reason))
        if not self.__connection.is_closed:
            self.__connection.close()

    def on_consumer_cancel_ok(self, _unused_frame, userdata):
        if not self.__is_consuming:
            self.__channel.close()

    def reconnect_token_expired(self, token):
        self.__token = token
        """
        Stop comsuming and invoke the stop function for channel and connection 
        """
        if self.__is_consuming:
            self.stop_consuming()
        """
        Check if the channel and connection are closed 
        """
        self.__until_all_closed_and_reconnect()

    async def __until_connected(self):
        if self.__channel is None or self.__connection is None:
            await asyncio.sleep(0.1)
        else:
            if self.__channel.is_open and self.__connection.is_open:
                return
            else:
                self.__loop.call_later(
                    0.1,
                    self.__until_connected
                )

    def __until_all_closed_and_reconnect(self):
        if not self.__channel.is_closed or not self.__connection.is_closed:
            self.__loop.call_later(
                0.1,
                self.__until_all_closed_and_reconnect
            )
        else:
            APP_LOGGER.info("[S3I]: Reconnect to Broker")
            self.connect()

    def send(self, endpoint, msg):
        if isinstance(msg, dict):
            msg = json.dumps(msg)
        if self.__channel.is_open:
            try:
                raise_error_from_s3ib_amqp(
                    self.__channel.basic_publish,
                    S3IBrokerAMQPError,
                    MSG_EXCHANGE,
                    endpoint,
                    msg,
                    pika.BasicProperties(
                        content_type="application/json",
                        delivery_mode=2
                    ))
            except S3IBrokerAMQPError as err:
                APP_LOGGER.error("[S3I]: Sending message failed: {}".format(err))
                return False

            else:
                return True

    def publish_event(self, msg, topic):
        if self.__channel.is_open:
            try:
                raise_error_from_s3ib_amqp(
                    self.__channel.basic_publish,
                    S3IBrokerAMQPError,
                    EVENT_EXCHANGE,
                    topic,
                    msg,
                    pika.BasicProperties(
                        content_type="application/json",
                    ))
            except S3IBrokerAMQPError as err:
                APP_LOGGER.error("[S3I]: Sending event failed: {}".format(err))
                return False

            else:
                return True



from ml.ml40.features.properties.values.value import Value


class LoadAlarm(Value):
    def __init__(self, namespace="mml40", name="", identifier=""):
        super().__init__(namespace=namespace, name=name, identifier=identifier)

        self.__is_alarm = None
        self.__displacement_threshold = None

    @property
    def isAlarm(self):
        return self.__is_alarm

    @isAlarm.setter
    def isAlarm(self, value):
        self.__is_alarm = value

    @property
    def displacementThreshold(self):
        return self.__displacement_threshold

    @displacementThreshold.setter
    def displacementThreshold(self, value):
        self.__displacement_threshold = value

    def to_json(self):
        self.__json_out = super().to_json()
        if self.__is_alarm is not None:
            self.__json_out["isAlarm"] = self.__is_alarm
        if self.__displacement_threshold is not None:
            self.__json_out["displacementThreshold"] = self.__displacement_threshold

        return self.__json_out



from ml.ml40.features.properties.values.value import Value


class ExpansionLength(Value):
    def __init__(self, namespace="ml40", name="", identifier="", parent=None):
        super().__init__(
            namespace=namespace, name=name, identifier=identifier, parent=parent
        )

        self.__current_length = None
        self.__max_length = None
        self.__json_out = dict()

    @property
    def currentLength(self):
        return self.__current_length

    @currentLength.setter
    def currentLength(self, value):
        self.__current_length = value

    @property
    def maxLength(self):
        return self.__max_length

    @maxLength.setter
    def maxLength(self, value):
        self.__max_length = value

    def to_json(self):
        self.__json_out = super().to_json()
        if self.__current_length is not None:
            self.__json_out["currentLength"] = self.__current_length
        if self.__max_length is not None:
            self.__json_out["maxLength"] = self.__max_length

        return self.__json_out



"""This module provides a collection of convenience functions."""

import json
import os
import xml.etree.ElementTree as ET
import sqlite3
import time
import jsonschema
from jsonschema import validate
from ml.app_logger import APP_LOGGER

IDENTITY_PROVIDER_URL = "https://idp.s3i.vswf.dev/"
BUILT_IN = ["from", "str", "dict", "list", "bool", "float", "int", "tuple",
               "def", "return", "for", "if", "class", "try", "except"]

class BColors:
    """colors for the console log"""

    HEADER = "\033[95m"
    OKBLUE = "\033[94m"
    OKGREEN = "\033[92m"
    WARNING = "\033[93m"
    FAIL = "\033[91m"
    ENDC = "\033[0m"
    BOLD = "\033[1m"
    UNDERLINE = "\033[4m"


def make_sub_thing(name, roles, features=[]):
    """
    Creates a dictionary representing this thing in it's current state
    as a subordinate thing. This representation should be used for
    subordinate things in s3i repository entries

    :param name: name of the sub thing
    :type name: str
    :param roles: fml40 roles of the sub thing
    :type roles: str
    :param features: fml40 features of the sub thing
    :type features: list

    :returns: JSON of the sub thing
    :rtype: dict

    """
    sub_thing = {
        "class": "ml40::Thing",
        "name": name,
        "roles": roles,
        "features": features
    }

    return sub_thing


def make_feature_config(class_name, identifier="", name="", subFeatures=""):
    """
    Creates a dictionary representing for its ml/fml40 features

    :param class_name: class name of the fml40 feature
    :type class_name: str
    :param identifier: local id of the fml40 feature
    :type identifier: str
    :param name: name of the fml40 feature
    :type name: str
    :param subFeatures: sub features
    :type subFeatures: dict

    :returns: JSON of the fml40 feature
    :rtype: dict

    """
    config_json = {
        "class": class_name
    }
    if identifier:
        config_json["identifier"] = identifier
    if name:
        config_json["name"] = name
    if subFeatures:
        config_json["subFeatures"] = subFeatures

    return config_json


def make_thing_config(thing_id, name, roles, features=[], config_path=""):
    """
    Creates a configuration file (JSON) for a fml40 thing.

    :param thing_id: identifier of the thing
    :type thing_id: str
    :param name: name of the thing
    :type name: str
    :param roles: fml40 roles of the thing
    :type roles: list
    :param features: fml40 features of the thing
    :type features: list
    :param config_path: path where the configuration file should be located
    :type config_path: str

    :returns: name of the configuration file of the thing
    :rtype: str


    """
    if not config_path:
        config_path = os.path.join("__file__", "configs")

    config_file = {
        "thingId": thing_id,
        "policyId": thing_id,
        "attributes": {
            "class": "ml40::Thing",
            "name": name,
            "roles": roles,
            "features": features,
        }
    }
    file_path = os.path.join(config_path, "{}.json".format(name))
    with open(file_path, 'wb') as file:
        file.write(json.dumps(config_file).encode('utf-8'))
    return "{}.json".format(name)


def load_config(config_filepath):
    """
    Loads a JSON object from a json formatted file found at config_filepath.

    :param config_filepath: Path to json formatted file.
    :type config_filepath: str

    """
    with open(config_filepath) as config_file:
        config = json.load(config_file)
        return config


def find_broker_endpoint(dir_obj, thing_id):
    """
    Finds the S3I-B endpoint of a thing

    :param dir_obj: S³I Directory Object
    :type dir_obj: object
    :param thing_id: identifier of the searched thing

    """
    thing_json = dir_obj.queryThingIDBased(thing_id)
    all_endpoints = thing_json["attributes"].get("allEndpoints", None)
    if all_endpoints:
        for ep in all_endpoints:
            if "s3ib" in ep:
                return ep


def remove_namespace(input_str):
    """
    Removes the specified namespace like ml40 or fml40

    :param input_str: input with namespace
    :type input_str: str
    :returns: output without namespace
    :rtype: str

    """
    return input_str.replace("mml40::", "").replace("fml40::", "").replace("ml40::", "")


def check_var_conflict(var):
    """
    check if the input is built-in variable
    :param var: python var
    :return: modified variable, if the input is a input variable.
    """

    if var in BUILT_IN:
        return "_{}".format(var)
    else:
        return var


class XML:
    def __init__(self, path=None, xml_str=None):
        self.et = ET
        self._namespace = {}

        self._namespace = {
            'xsi': 'http://www.w3.org/2001/XMLSchema-instance',
            'schemaLocation': 'urn:skogforsk:stanford2010 HarvestedProduction_V3p0.xsd',
            '': 'urn:skogforsk:stanford2010'
        }

        for key in self._namespace:
            self.et.register_namespace(key, self._namespace[key])

        if path is not None:
            self.tree = self.et.parse(path)
            self.root = self.tree.getroot()

        if xml_str is not None:
            self.root = self.et.fromstring(xml_str)

    def find_nodes(self, path):
        path_list = path.split('/')
        new_path = ""
        for _path in path_list:
            if ":" not in _path:
                _path = "{" + self._namespace.get('') + "}" + _path
            new_path += _path
            new_path += "/"
        if new_path[-1] == "/":
            new_path = new_path[:-1]
        return self.root.findall(new_path, self._namespace)

    def to_string(self, root):
        return self.et.tostring(root, "unicode", "xml")


class DataBase:
    def __init__(self, db):
        self.__db = db
        self.__table_conf = {}
        self.__conn = None

    @property
    def conn(self):
        return self.__conn

    @property
    def cursor(self):
        return self.__conn.cursor()

    def connect(self):
        self.__conn = sqlite3.connect(self.__db, check_same_thread=False)
        APP_LOGGER.info("Connect to the database {}".format(self.__db))

    def build_table(self, table, configuration):
        conf_schema = {
            "type": "array",
            "items": {
                "type": "object",
                "required": ["var_name", "var_type"],
                "properties": {
                    "var_name": {"type": "string"},
                    "var_type": {"type": "string"}
                }
            }
        }
        try:
            validate(instance=configuration, schema=conf_schema)
        except jsonschema.ValidationError as e:
            APP_LOGGER.error("config json error: {}".format(e))

        exec_str = "ID INTEGER PRIMARY KEY AUTOINCREMENT"

        for i in configuration:
            name = i["var_name"]
            type = i["var_type"]
            exec_str += ", {} {}".format(name, type)

        exec_str = '''CREATE TABLE if not exists {0} ({1});'''.format(table, exec_str)
        cursor = self.__conn.cursor()
        try:
            cursor.execute(exec_str)
            APP_LOGGER.info("Create table {} if not exists".format(table))
        except sqlite3.OperationalError as e:
            APP_LOGGER.warning(e)

        self.__table_conf[table] = configuration

    def disconnect(self):
        self.__conn.close()

    def update(self, table, value_dict):
        for key in value_dict.keys():
            def_var_name = []
            for var in self.__table_conf.get(table, {}):
                def_var_name.append(var["var_name"])
            if key not in def_var_name:
                raise ValueError("{} is not defined variable name".format(key))

        cursor = self.__conn.cursor()
        exec_str = ""

        for i in value_dict.keys():
            exec_str += "{},".format(i)
        """    
        for i in self.__table_conf.get(table):
            exec_str += "{},".format(i["var_name"])
        """
        exec_str = exec_str[:-1]
        exec_str = "INSERT INTO {} ({})".format(table, exec_str)

        value_str = ""

        for i in value_dict.keys():
            if isinstance(value_dict[i], str):
                value_str += "'{}',".format(value_dict[i])
            elif isinstance(value_dict[i], (float, int, bool)):
                value_str += "{},".format(value_dict[i])
        """        
        for i in self.__table_conf.get(table, {}):
            if isinstance(value_dict[i["var_name"]], str):
                value_str += "'{}',".format(value_dict[i["var_name"]])
            elif isinstance(value_dict[i["var_name"]], (float, int, bool)):
                value_str += "{},".format(value_dict[i["var_name"]])
        """
        value_str = value_str[:-1]
        value_str = "VALUES ({})".format(value_str)

        exec_str += value_str
        try:
            cursor.execute(exec_str)
        except sqlite3.OperationalError as e:
            APP_LOGGER.warning(e)
        self.__conn.commit()

    def delete(self, table, id=None):
        if id is None:
            exec_str = "DELETE FROM {};".format(table)
        else:
            exec_str = "DELETE FROM {} where id={}".format(table, id)
        try:
            self.__conn.cursor().execute(exec_str)
        except sqlite3.OperationalError as e:
            APP_LOGGER.warning(e)

        self.__conn.commit()

    def drop(self, table):
        exec_str = "DROP TABLE {}".format(table)
        try:
            self.__conn.cursor().execute(exec_str)
        except sqlite3.OperationalError as e:
            APP_LOGGER.warning(e)
        self.__conn.commit()

    def search(self, table):
        exec_str = "SELECT * FROM {};".format(table)
        cursor = self.__conn.cursor().execute(exec_str)
        for row in cursor:
            print("------------------")
            print("ID = {}".format(row[0]))
            for i in range(len(self.__table_conf.get(table, {}))):
                print("{} = {}".format(self.__table_conf.get(table)[i]["var_name"].upper(), row[i + 1]))
            print("------------------\n\n")

    def execute(self, exec_str):
        try:
            return self.__conn.cursor().execute(exec_str)
        except sqlite3.OperationalError as e:
            APP_LOGGER.warning(e)




from ml.ml40.features.properties.values.value import Value


class HighSpeedImageValue(Value):
    def __init__(self, namespace="ml40", name="", identifier=""):
        super().__init__(
            namespace=namespace,
            name=name,
            identifier=identifier)

        self.__number_of_pixel = None
        self.__min_range = None
        self.__max_range = None

        self.__min_pixel_value = None
        self.__max_pixel_value = None

        self.__mean_pixel_value = None

    @property
    def numberOfPixel(self):
        return self.__number_of_pixel

    @numberOfPixel.setter
    def numberOfPixel(self, value):
        self.__number_of_pixel = value

    @property
    def minRange(self):
        return self.__min_range

    @minRange.setter
    def minRange(self, value):
        self.__min_range = value

    @property
    def maxPixelValue(self):
        return self.__max_pixel_value

    @maxPixelValue.setter
    def maxPixelValue(self, value):
        self.__max_pixel_value = value

    @property
    def minPixelValue(self):
        return self.__min_pixel_value

    @minPixelValue.setter
    def minPixelValue(self, value):
        self.__min_pixel_value = value

    @property
    def meanPixelValue(self):
        return self.__mean_pixel_value

    @meanPixelValue.setter
    def meanPixelValue(self, value):
        self.__mean_pixel_value = value

    def to_json(self):
        self.__json_out = super().to_json()
        if self.__number_of_pixel is not None:
            self.__json_out["numberOfPixel"] = self.__number_of_pixel

        if self.__min_range is not None:
            self.__json_out["minRange"] = self.__min_range

        if self.__max_range is not None:
            self.__json_out["maxRange"] = self.__max_range

        if self.__min_pixel_value is not None:
            self.__json_out["minPixelValue"] = self.__min_pixel_value

        if self.__max_pixel_value is not None:
            self.__json_out["maxPixelValue"] = self.__max_pixel_value

        if self.__mean_pixel_value is not None:
            self.__json_out["meanPixelValue"] = self.__mean_pixel_value

        return self.__json_out




from ml.ml40.features.functionalities.functionality import Functionality
from ml.ml40.features.properties.values.documents.reports.report import Report


class ProvidesOperationalData(Functionality):
    def __init__(self, namespace="ml40", name="", identifier="", parent=None):
        super().__init__(
            namespace=namespace, name=name, identifier=identifier, parent=parent
        )

    def getOperationalData(self) -> Report:
        return "this is a simply report"



""" This module implements a factory for managing and creating Digital Twins according to Forest Modeling Language 4.0."""

from ml.app_logger import APP_LOGGER
from ml.tools import remove_namespace, check_var_conflict
from ml.ditto_feature import ditto_feature
import sys
import inspect
from ml.parameters import Parameters
from ml.entry import Entry
from ml.thing import Thing
from ml.ml40.roles.servives.service import Service
from ml.ml40.roles.hmis.app import App
from ml.ml40.roles.hmis.dashboard import Dashboard
from ml.ml40.roles.hmis.machine_ui import MachineUI
from ml.ml40.roles.hmis.hmd import HMD
from ml.ml40.roles.hmis.hmi import HMI
from ml.ml40.roles.dts.handheld_devices.handheld_device import HandheldDevice
from ml.ml40.roles.dts.machines.machine import Machine
from ml.ml40.roles.dts.parts.crane import Crane
from ml.ml40.roles.dts.parts.part import Part
from ml.ml40.roles.dts.parts.engine import Engine
from ml.ml40.roles.dts.parts.scale import Scale
from ml.ml40.roles.dts.parts.tank import Tank
from ml.ml40.roles.dts.persons.machine_operator import MachineOperator
from ml.ml40.roles.dts.persons.person import Person
from ml.ml40.roles.dts.sensors.sensor import Sensor
from ml.ml40.roles.dts.sensors.air_sensor import AirSensor
from ml.ml40.roles.dts.sensors.soil_sensor import SoilSensor
from ml.ml40.roles.dts.sites.site import Site
from ml.ml40.roles.dts.ways.way import Way

from ml.fml40.roles.dts.handheld_devices.brushcutter import Brushcutter
from ml.fml40.roles.dts.handheld_devices.chainsaw import Chainsaw
from ml.fml40.roles.dts.machines.forest_machine import ForestMachine
from ml.fml40.roles.dts.machines.forwarder import Forwarder
from ml.fml40.roles.dts.machines.harvester import Harvester
from ml.fml40.roles.dts.machines.log_truck import LogTruck
from ml.fml40.roles.dts.machines.mini_tractor import MiniTractor
from ml.fml40.roles.dts.machines.skidder import Skidder
from ml.fml40.roles.dts.machines.wheel_loader import WheelLoader
from ml.fml40.roles.dts.parts.grabber import Grabber
from ml.fml40.roles.dts.parts.harvesting_head import HarvestingHead
from ml.fml40.roles.dts.parts.log_loading_area import LogLoadingArea
from ml.fml40.roles.dts.parts.saw import Saw
from ml.fml40.roles.dts.parts.stacking_shield import StackingShield
from ml.fml40.roles.dts.parts.winch import Winch
from ml.fml40.roles.dts.persons.forest_owner import ForestOwner
from ml.fml40.roles.dts.persons.forest_worker import ForestWorker
from ml.fml40.roles.dts.persons.mini_tractor_operator import MiniTractorOperator
from ml.fml40.roles.dts.persons.skidder_operator import SkidderOperator
from ml.fml40.roles.dts.sensors.vitality_sensor import VitalitySensor
from ml.fml40.roles.dts.sensors.barkbeetle_sensor import BarkbeetleSensor
from ml.fml40.roles.dts.sites.forest_enterprise import ForestEnterprise
from ml.fml40.roles.dts.sites.hauler import Hauler
from ml.fml40.roles.dts.sites.mill.mill import Mill
from ml.fml40.roles.dts.sites.mill.papermill import Papermill
from ml.fml40.roles.dts.sites.mill.sawmill import Sawmill
from ml.fml40.roles.dts.woods.wood import Wood
from ml.fml40.roles.dts.woods.stem_segment import StemSegment
from ml.fml40.roles.dts.woods.wood_pile import WoodPile
from ml.fml40.roles.dts.forest.forest import Forest
from ml.fml40.roles.dts.forest.forest_segment import ForestSegment
from ml.fml40.roles.dts.forest.tree import Tree

from ml.mml40.roles.dts.parts.cantilever import Cantilever
from ml.mml40.roles.dts.sensors.strain_gauge import StrainGauge

from ml.ml40.features.properties.associations.association import Association
from ml.ml40.features.properties.associations.composite import Composite
from ml.ml40.features.properties.property import Property
from ml.ml40.features.properties.associations.shared import Shared
from ml.ml40.features.properties.values.address import Address
from ml.ml40.features.properties.values.count import Count
from ml.ml40.features.properties.values.dimensions import Dimensions
from ml.ml40.features.properties.values.distance import Distance
from ml.ml40.features.properties.values.expansion_length import ExpansionLength
from ml.ml40.features.properties.values.force import Force
from ml.ml40.features.properties.values.generic_property import GenericProperty
from ml.ml40.features.properties.values.last_service_check import LastServiceCheck
from ml.ml40.features.properties.values.lot import Lot
from ml.ml40.features.properties.values.lift import Lift
from ml.ml40.features.properties.values.liquid_filling_level import LiquidFillingLevel
from ml.ml40.features.properties.values.location import Location
from ml.ml40.features.properties.values.moisture import Moisture
from ml.ml40.features.properties.values.operating_hours import OperatingHours
from ml.ml40.features.properties.values.orientation_rpy import OrientationRPY
from ml.ml40.features.properties.values.personal_name import PersonalName
from ml.ml40.features.properties.values.rotational_speed import RotationalSpeed
from ml.ml40.features.properties.values.route import Route
from ml.ml40.features.properties.values.switching_stage import SwitchingStage
from ml.ml40.features.properties.values.temperature import Temperature
from ml.ml40.features.properties.values.time_slot import TimeSlot
from ml.ml40.features.properties.values.weight import Weight
from ml.ml40.features.properties.values.documents.jobs.generic_job import GenericJob
from ml.ml40.features.properties.values.documents.jobs.job import Job
from ml.ml40.features.properties.values.documents.jobs.job_list import JobList
from ml.ml40.features.properties.values.documents.jobs.job_status import JobStatus
from ml.ml40.features.properties.values.documents.reports.report import Report
from ml.ml40.features.properties.values.documents.reports.production_data import ProductionData

from ml.fml40.features.properties.values.abstract_inventory import AbstractInventory
from ml.fml40.features.properties.values.assortment import Assortment
from ml.fml40.features.properties.values.dbh import DBH
from ml.fml40.features.properties.values.fell_indicator import FellIndicator
from ml.fml40.features.properties.values.felling_period import FellingPeriod
from ml.fml40.features.properties.values.harvesting_parameter import HarvestingParameters
from ml.fml40.features.properties.values.harvested_volume import HarvestedVolume
from ml.fml40.features.properties.values.interfering_branches import InterferingBranches
from ml.fml40.features.properties.values.inventory_data import InventoryData
from ml.fml40.features.properties.values.overhang import Overhang
from ml.fml40.features.properties.values.stem_segment_properties import StemSegmentProperties
from ml.fml40.features.properties.values.thickness_class import ThicknessClass
from ml.fml40.features.properties.values.tilt import Tilt
from ml.fml40.features.properties.values.tree_data import TreeData
from ml.fml40.features.properties.values.tree_type import TreeType
from ml.fml40.features.properties.values.vitality_status import VitalityStatus
from ml.fml40.features.properties.values.wood_quality import WoodQuality

from ml.fml40.features.properties.values.documents.jobs.felling_job import FellingJob
from ml.fml40.features.properties.values.documents.jobs.fellung_support_job import FellingSupportJob
from ml.fml40.features.properties.values.documents.jobs.forwarding_job import ForwardingJob
from ml.fml40.features.properties.values.documents.jobs.log_transportation_job import LogTransportationJob

from ml.fml40.features.properties.values.documents.reports.afforestation_suggestion import AfforestationSuggestion
from ml.fml40.features.properties.values.documents.reports.felling_tool import FellingTool
from ml.fml40.features.properties.values.documents.reports.log_measurement import LogMeasurement
from ml.fml40.features.properties.values.documents.reports.log_transportation_report import LogTransportationReport
from ml.fml40.features.properties.values.documents.reports.map_data import MapData
from ml.fml40.features.properties.values.documents.reports.moisture_prediction_report import MoisturePredictionReport
from ml.fml40.features.properties.values.documents.reports.passability_report import PassabilityReport
from ml.fml40.features.properties.values.documents.reports.soil_moisture_measurement import SoilMoistureMeasurement

from ml.mml40.features.properties.values.Displacement import Displacement
from ml.mml40.features.properties.values.GeometryProperties import GeometryProperties
from ml.mml40.features.properties.values.LoadAlarm import LoadAlarm
from ml.mml40.features.properties.values.MaterialProperties import MaterialProperties
from ml.mml40.features.properties.values.Stretch import Stretch


from ml.ml40.features.functionalities.accepts_jobs import AcceptsJobs
from ml.ml40.features.functionalities.accepts_reports import AcceptsReports
from ml.ml40.features.functionalities.clears_jobs import ClearsJobs
from ml.ml40.features.functionalities.functionality import Functionality
from ml.ml40.features.functionalities.manages_jobs import ManagesJobs
from ml.ml40.features.functionalities.plans_routes import PlansRoutes
from ml.ml40.features.functionalities.provides_map_data import ProvidesMapData
from ml.ml40.features.functionalities.provides_operational_data import ProvidesOperationalData
from ml.ml40.features.functionalities.renders import Renders

from ml.fml40.features.functionalities.accepts_felling_jobs import AcceptsFellingJobs
from ml.fml40.features.functionalities.accepts_felling_support_jobs import AcceptsFellingSupportJobs
from ml.fml40.features.functionalities.accepts_forwarding_jobs import AcceptsForwardingJobs
from ml.fml40.features.functionalities.accepts_log_measurements import AcceptsLogMeasurements
from ml.fml40.features.functionalities.accepts_log_transportaition_jobs import AcceptsLogTransportationJobs
from ml.fml40.features.functionalities.accepts_moisture_measurement import AcceptsMoistureMeasurement
from ml.fml40.features.functionalities.accepts_move_commands import AcceptsMoveCommands
from ml.fml40.features.functionalities.accepts_passability_report import AcceptsPassabilityReport
from ml.fml40.features.functionalities.accepts_proximity_alert import AcceptsProximityAlert
from ml.fml40.features.functionalities.accepts_shield_commands import AcceptsShieldCommands
from ml.fml40.features.functionalities.accepts_single_tree_felling_jobs import AcceptsSingleTreeFellingJobs
from ml.fml40.features.functionalities.accepts_winch_command import AcceptsWinchCommands
from ml.fml40.features.functionalities.classifies_tree_species import ClassifiesTreeSpecies
from ml.fml40.features.functionalities.cuts import Cuts
from ml.fml40.features.functionalities.determines_passability import DeterminesPassability
from ml.fml40.features.functionalities.displays_health_alarms import DisplaysHealthAlarms
from ml.fml40.features.functionalities.evaluates_stand_attributes import EvaluatesStandAttributes
from ml.fml40.features.functionalities.fells import Fells
from ml.fml40.features.functionalities.forest_planning_evaluation import ForestPlanningEvaluation
from ml.fml40.features.functionalities.forwards import Forwards
from ml.fml40.features.functionalities.generates_afforestation_suggestions import GeneratesAfforestationSuggestions
from ml.fml40.features.functionalities.generates_felling_suggestions import GeneratesFellingSuggestions
from ml.fml40.features.functionalities.grabs import Grabs
from ml.fml40.features.functionalities.harvests import Harvests
from ml.fml40.features.functionalities.measure_wood import MeasuresWood
from ml.fml40.features.functionalities.monitor_health_status import MonitorsHealthStatus
from ml.fml40.features.functionalities.provides_moisture_prediction import ProvidesMoisturePrediction
from ml.fml40.features.functionalities.provides_passability_information import ProvidesPassabilityInformation
from ml.fml40.features.functionalities.provides_production_data import ProvidesProductionData
from ml.fml40.features.functionalities.provides_tree_data import ProvidesTreeData
from ml.fml40.features.functionalities.provides_weather_data import ProvidesWeatherData
from ml.fml40.features.functionalities.simulates_tree_growth import SimulatesTreeGrowth
from ml.fml40.features.functionalities.supports_felling import SupportsFelling
from ml.fml40.features.functionalities.transports_logs import TransportsLogs

from ml.mml40.features.functionalities.CantileverConfigure import CantileverConfigure
from ml.mml40.features.functionalities.ProvidesDisplacementData import ProvidesDisplacementData
from ml.mml40.features.functionalities.ProvidesForceData import ProvidesForceData
from ml.mml40.features.functionalities.ProvidesStretchData import ProvidesStretchData
# TODO: Get rid of this global variable
# TODO: automatically get all classes in module
DT_FACTORY = {}

clsmembers = inspect.getmembers(sys.modules[__name__], inspect.isclass)
for member in clsmembers:
    DT_FACTORY[member[0]] = member[1]


def build_sub_features(feature_ins, feature):
    """
    Instantiates and inserts ml40/fml40 sub features object into feature instance

    :param feature_ins: ml40/fml40 feature instance, which has a sub feature to be built
    :type feature_ins: object
    :param feature: ml40/fml40 feature containing subFeatures
    :type feature: dict

    """
    sub_features = feature.get("subFeatures", [])
    for sub_f in sub_features:
        sub_f_name = sub_f.get("name")
        if sub_f_name is None:
            sub_f_name = sub_f.get("identifier")
        if sub_f_name is None:
            sub_f_name = sub_f.get("class")
        _class_name = sub_f.get("class", "")
        sub_f_obj = DT_FACTORY.get(remove_namespace(_class_name), None)
        if sub_f_obj is None:
            APP_LOGGER.critical("Subfeature: %s is missing" % _class_name)
        else:
            APP_LOGGER.info("Adding subfeature: %s" % _class_name)
            sub_f_instance = sub_f_obj()
            for key in sub_f.keys():
                if key == "targets":
                    build_sub_thing(sub_f_instance, sub_f)
                elif key == "subFeatures":
                    build_sub_features(sub_f_instance, sub_f)
                else:
                    setattr(sub_f_instance, key, sub_f[key])
            sub_f_instance.parent = feature_ins
            feature_ins.subFeatures[sub_f_name] = sub_f_instance


def build_sub_thing(feature_ins, feature):
    """
    Instantiates and inserts sub thing object into a feature instance

    :param feature_ins: ml40/fml40 feature instance
    :type feature_ins: object
    :param feature: ml40/fml40 feature, which contains a sub thing
    :type feature: dict

    """
    json_sub_things = feature.get("targets", [])
    for json_sub_thing in json_sub_things:
        sub_thing_ref = create_thing(model_json={"attributes": json_sub_thing})
        sub_thing_name = json_sub_thing.get("name", None)
        feature_ins.targets[sub_thing_name] = sub_thing_ref
        sub_thing_ref.parent = feature_ins


def build(thing, model):
    """
    Builds a ml40 thing instance

    :param thing: ml40 thing instance
    :type thing: object
    :param model: model in ml40 thing JSON
    :type model: dict

    """
    attributes = model.get("attributes", None)
    ditto_features = model.get("features", None)

    if not isinstance(model, dict):
        # TODO JSON Schema
        APP_LOGGER.critical("model is no valid JSON")
        return
    roles = attributes.get("roles", [])
    for role in roles:
        role_instance = build_role(role)
        role_instance.parent = thing
        thing.roles[role.get("class")] = role_instance

    json_features = attributes.get("features", [])
    for feature in json_features:
        feature_ins = build_feature(feature=feature)
        feature_ins.parent = thing
        thing.features[feature.get("class")] = feature_ins

    if ditto_features is not None:
        build_ditto_features(thing, ditto_features)


def build_ditto_features(thing, ditto_features):
    for id in ditto_features.keys():
        for key in ditto_features[id]["properties"]:
            ditto_f = ditto_feature(id=id, key=key, value=ditto_features[id]["properties"][key])
        thing.ditto_features[ditto_f.id] = ditto_f


def build_role(role):
    """
     Instantiates and inserts a ml40/fml40 role object into a ml40 thing instance

    :param role: ml40/fml40 role
    :type role: dict

    """
    role_class_name = role.get("class", "")
    role_obj = DT_FACTORY.get(remove_namespace(role_class_name), None)
    if role_obj is None:
        APP_LOGGER.critical("Roles: %s is missing" % role_class_name)
        role_instance = None
    else:
        APP_LOGGER.info("Adding roles: %s" % role_class_name)
        role_instance = role_obj()
    return role_instance


def build_feature(feature):
    """
     Instantiates and inserts a ml40/fml40 feature object in a ml40 thing instance

    :param feature: ml40/fml40 feature
    :type feature: dict

    """
    feature_class_name = feature.get("class", "")
    feature_obj = DT_FACTORY.get(remove_namespace(feature_class_name), None)

    if feature_obj is None:
        APP_LOGGER.critical("Feature: %s is missing" % feature_class_name)
        feature_instance = None
    else:
        APP_LOGGER.info("Adding feature: %s" % feature_class_name)
        feature_instance = feature_obj()
        for key in feature.keys():
            if key == "class":
                continue
            if key == "targets":
                build_sub_thing(feature_instance, feature)
            elif key == "subFeatures":
                build_sub_features(feature_instance, feature)
            else:
                setattr(feature_instance, check_var_conflict(key), feature[key])
    return feature_instance


def add_function_impl_obj(thing, impl_obj, feature_name, **kwargs):
    """
    Adds user-specific implemented object to a thing instance

    :param thing: ml40 thing instance
    :type thing: object
    :param impl_obj: ml40/fml40 feature instance
    :type impl_obj: object
    :param feature_name: class name of a ml40/fml40 feature
    :type feature_name: str


    """
    feature = thing.features.get(feature_name, None)
    if feature is None:
        ### Check if it is a feature of a subthing
        if thing.features.get("ml40::Composite", None) is not None:
            subthings = thing.features["ml40::Composite"].targets
            for i in subthings.keys():
                feature = subthings[i].features.get(feature_name, None)
                if feature is not None:
                    break
    if feature is None:
        if thing.features.get("ml40::Shared", None) is not None:
            subthings = thing.features["ml40::Shared"].targets
            for i in subthings.keys():
                feature = subthings[i].features.get(feature_name, None)
                if feature is not None:
                    break
    if feature is None:
        APP_LOGGER.critical(
            "Functionality %s is not one of the build-in functionalities" % feature_name
        )
    else:
        APP_LOGGER.info("Implementation object is added into the functionality %s" % feature_name)
        impl_ins = impl_obj(**kwargs)
        impl_ins.class_name = feature_name
        thing.features[feature_name] = impl_ins


def create_thing(
        model_json,
        oauth2_secret="",
        grant_type="client_credentials",
        username=None,
        password=None,
        is_repository=False,
        is_broker=False,
        is_broker_rest=False,
        parameters=Parameters()
):
    attributes = model_json.get("attributes")
    if attributes is None:
        APP_LOGGER.error("Attributes are empty")

    identifier = model_json.get("thingId", "")
    name = attributes.get("name", "")
    APP_LOGGER.info("Build digital twin {} with id {}".format(name, identifier))
    entry_ref = Entry(identifier, name)
    build(entry_ref, model_json)
    return Thing(
        entry=entry_ref,
        oauth2_secret=oauth2_secret,
        grant_type=grant_type,
        username=username,
        password=password,
        is_repository=is_repository,
        is_broker=is_broker,
        is_broker_rest=is_broker_rest,
        parameters=parameters
    )


def _create_thing(model, grant_type="password",
                 secret="", username=None, password=None,
                 is_broker_rest=False, is_broker=False, is_repo=False, is_stanford2010=False,
                 is_stanford2010_sync=False,
                 stanford2010_sync_freq=None, stanford2010_path=None, is_database=False,
                 database_conf=None, database_file=None):
    """
    Creates and launches a thing which connects to the S3I

    :param model: edge-device or S³I Repository specified JSON entry
    :type model: dict
    :param grant_type: grant type of OAuth2.0, which can be password or client_credentials
    :type grant_type: str
    :param secret: secret of a thing
    :type secret: str
    :param username: username, if the grant_type is set as password, the username is required
    :type username: str
    :param password: password, if the grant_type is set as password, the password is required
    :type password: str
    :param is_broker: whether broker interface is enabled in the ml40::thing instance
    :type is_broker: bool
    :param is_broker_rest: whether the broker interface uses the HTTP REST
    :type is_broker_rest: bool
    :param is_repo: whether the repository interface is enabled in the ml40::thing instance
    :type is_repo: bool

    :returns: ml40::thing instance
    :rtype: object

    """
    attributes = model.get("attributes", None)
    if attributes is None:
        sys.exit("attributes are none")
    thing_type = attributes.get("class", "")
    thing_type = remove_namespace(thing_type)
    thing_name = attributes.get("name", "")
    APP_LOGGER.info("Build digital twin {} with id {}".format(thing_name, model.get("thingId", "")))

    _thing = DT_FACTORY.get(thing_type)

    thing_ref = _thing(
        model=model,
        grant_type=grant_type,
        client_secret=secret,
        username=username,
        password=password,
        is_broker_rest=is_broker_rest,
        is_broker=is_broker,
        is_repo=is_repo,
        is_stanford2010=is_stanford2010,
        stanford2010_sync_freq=stanford2010_sync_freq,
        is_stanford2010_sync=is_stanford2010_sync,
        stanford2010_path=stanford2010_path,
        is_database=is_database,
        database_file=database_file,
    )
    build(thing_ref, model)

    return thing_ref
